import pathlib
import os
from scipy.spatial.transform import Rotation
import numpy as np
import healpix as hp
import logging


def angle_to_angle_list(
    angle_diff: float, sort_angles: bool = True, log_level: int = logging.DEBUG
) -> list[tuple[float, float, float]]:
    """Auto generate an angle list for a given maximum angle difference.

    The code uses healpix to determine Z1 and X and splits Z2 linearly.

    Parameters
    ----------
    angle_diff: float
        maximum difference (in degrees) for the angle list
    sort_angles: bool, default True
        sort the list, using python default angle_list.sort(), sorts first on Z1,
        then X, then Z2
    log_level: int, default logging.DEBUG
        the log level to use when generating logs



    Returns
    -------
    angle_list: list[tuple[float, float, float]]
        a list where each element is a tuple of 3 floats containing
        an anti-clockwise ZXZ Euler rotation in radians
    """
    # We use an approximation of the square root of the area as the median angle diff
    # This works reasonably well and is based on the following formula:
    # angle_diff = (4*np.pi/npix)**0.5 * 360/(2*np.pi)
    npix = 4 * np.pi / (angle_diff * np.pi / 180) ** 2
    nside = 0
    while hp.nside2npix(nside) < npix:
        nside += 1
    used_npix = hp.nside2npix(nside)
    used_angle_diff = (4 * np.pi / used_npix) ** 0.5 * (180 / np.pi)
    logging.log(
        log_level, f"Using an angle difference of {used_angle_diff:.4f} for Z1 and X"
    )
    theta, phi = hp.pix2ang(nside, np.arange(used_npix))
    # Now for psi
    n_psi_angles = int(np.ceil(360 / angle_diff))
    psi, used_psi_diff = np.linspace(
        0, 2 * np.pi, n_psi_angles, endpoint=False, retstep=True
    )
    logging.log(
        log_level,
        f"Using an angle difference of {np.rad2deg(used_psi_diff):.4f} for Z2",
    )
    angle_list = [(ph, th, ps) for ph, th in zip(phi, theta) for ps in psi]
    if sort_angles:
        angle_list.sort()
    return angle_list


def load_angle_list(
    file_name: pathlib.Path, sort_angles: bool = True
) -> list[tuple[float, float, float]]:
    """Load an angular search list from disk.

    Parameters
    ----------
    file_name: pathlib.Path
        path to text file containing angular search, each line should contain 3 floats
        of anti-clockwise ZXZ
    sort_angles: bool, default True
        sort the list, using python default angle_list.sort(), sorts first on Z1,
        then X, then Z2

    Returns
    -------
    angle_list: list[tuple[float, float, float]]
        a list where each element is a tuple of 3 floats containing an anti-clockwise
        ZXZ Euler rotation in radians
    """
    with open(str(file_name)) as fstream:
        lines = fstream.readlines()
    angle_list = [tuple(map(float, x.strip().split(" "))) for x in lines]
    if not all([len(a) == 3 for a in angle_list]):
        raise ValueError(
            "Invalid angle file provided, each line should have 3 ZXZ Euler angles!"
        )
    if sort_angles:
        # angle_list needs to be sorted otherwise symmetry reduction cannot be used!
        angle_list.sort()
    return angle_list


def get_angle_list(
    angle: pathlib.Path | float,
    sort_angles: bool = True,
    symmetry: int = 1,
    log_level: int = logging.DEBUG,
):
    """Either get an angular search file from disk or generate one from a float

    Parameters
    ----------
    angle: Union[pathlib.Path, float]
        either the path to text file containing angular search,
          each line should contain 3 floats of anti-clockwise ZXZ
        or if a float:
          maximum difference (in degrees) for the angle list
    sort_angles: bool, default True
        sort the list, using python default angle_list.sort(), sorts first on Z1,
        then X, then Z2
    symmetry: int, default 1
        the returned list will only have Z2 angles [0, (2*pi/symmetry))
    log_level: int, default logging.DEBUG
        the log level to use when generating logs

    Returns
    -------
    angle_list: list[tuple[float, float, float]]
        a list where each element is a tuple of 3 floats containing an anti-clockwise
        ZXZ Euler rotation in radians
    """
    out = None
    max_z2 = 2 * np.pi / symmetry
    try:
        angle = float(angle)
        angle_is_float = True
    except (ValueError, TypeError):
        angle_is_float = False
    if angle_is_float:
        logging.log(
            log_level,
            f"Will generate an angle list with a maximum increment of {angle}",
        )
        out = angle_to_angle_list(angle, sort_angles, log_level)
    elif isinstance(angle, str | os.PathLike):
        possible_file_path = pathlib.Path(angle)
        if possible_file_path.exists() and possible_file_path.suffix == ".txt":
            logging.log(
                log_level,
                "Custom file provided for the angular search. "
                "Checking if it can be read...",
            )
            out = load_angle_list(angle, sort_angles)

    if out is None:  # If no angles by now, error out
        raise ValueError("Invalid angle input provided")
    return [i for i in out if i[2] < max_z2]


def convert_euler(
    angles: tuple[float, float, float],
    order_in: str = "ZXZ",
    order_out: str = "ZXZ",
    degrees_in: bool = True,
    degrees_out: bool = True,
) -> tuple[float, float, float]:
    """Convert a single set of Euler angles from one Euler notation to another. This
    function makes use of scipy.spatial.transform.Rotation meaning that capital letters
    (i.e. ZXZ) specify intrinsic rotations (commonly used in cryo-EM) and small letters
    (i.e. zxz) specific extrinsic rotations.

    Parameters
    ----------
    angles: tuple[float, float, float]
        tuple of three angles
    order_in: str, default 'ZXZ'
        Euler rotation axis of input angles
    order_out: str, default 'ZXZ'
        Euler rotation axis of output angles
    degrees_in: bool, default True
        whether the input angles are in degrees
    degrees_out: bool, default True
        whether the output angles should be in degrees

    Returns
    -------
    output: tuple[float, float, float]
        tuple of three angles
    """
    r = Rotation.from_euler(order_in, angles, degrees=degrees_in)
    return tuple(r.as_euler(order_out, degrees=degrees_out))



# This file is named testzz_* as it should run last,
# because it permanently destroys the imports
# No imports of pytom_tm outside of the methods
import unittest
from importlib import reload

# Mock out installed dependencies
orig_import = __import__

# skip tests if optional stuff is not installed
SKIP_PLOT = False
try:
    import pytom_tm.plotting  # noqa: F401
except RuntimeError:
    SKIP_PLOT = True


def module_not_found_mock(missing_name):
    def import_mock(name, *args):
        if name == missing_name:
            raise ModuleNotFoundError(f"No module named '{name}'")
        return orig_import(name, *args)

    return import_mock


def cupy_import_error_mock(name, *args):
    if name == "cupy":
        raise ImportError("Failed to import cupy")
    return orig_import(name, *args)


class TestMissingDependencies(unittest.TestCase):
    def test_missing_cupy(self):
        # assert working import
        with self.assertNoLogs(level="WARNING"):
            import pytom_tm
        cupy_not_found = module_not_found_mock("cupy")
        # Test missing cupy
        with unittest.mock.patch("builtins.__import__", side_effect=cupy_not_found):
            with self.assertLogs(level="WARNING") as cm:
                reload(pytom_tm)
            self.assertEqual(len(cm.output), 1)
            self.assertIn("cupy installation not found or not functional", cm.output[0])

    def test_broken_cupy(self):
        # assert working import
        with self.assertNoLogs(level="WARNING"):
            import pytom_tm
        # Test cupy ImportError
        with unittest.mock.patch(
            "builtins.__import__", side_effect=cupy_import_error_mock
        ):
            with self.assertLogs(level="WARNING") as cm:
                reload(pytom_tm)
            self.assertEqual(len(cm.output), 1)
            self.assertIn("cupy installation not found or not functional", cm.output[0])

    @unittest.skipIf(SKIP_PLOT, "plotting module not installed")
    def test_missing_matplotlib(self):
        # assert working import
        import pytom_tm

        matplotlib_not_found = module_not_found_mock("matplotlib.pyplot")
        with unittest.mock.patch(
            "builtins.__import__", side_effect=matplotlib_not_found
        ):
            with self.assertRaisesRegex(ModuleNotFoundError, "matplotlib"):
                # only pyplot is directly imported so this should be tested
                import matplotlib.pyplot as plt  # noqa: F401
            # force reload
            # check if we can still import pytom_tm
            reload(pytom_tm)

            # check if plotting is indeed disabled after reload
            # (reload is needed to prevent python import caching)
            self.assertFalse(reload(pytom_tm.extract).plotting_available)
            # assert that importing the plotting module fails completely
            with self.assertRaisesRegex(RuntimeError, "matplotlib and seaborn"):
                reload(pytom_tm.plotting)

    @unittest.skipIf(SKIP_PLOT, "plotting module not installed")
    def test_missing_seaborn(self):
        # assert working import
        import pytom_tm

        seaborn_not_found = module_not_found_mock("seaborn")
        with unittest.mock.patch("builtins.__import__", side_effect=seaborn_not_found):
            with self.assertRaisesRegex(ModuleNotFoundError, "seaborn"):
                import seaborn  # noqa: F401
            # check if we can still import pytom_tm
            reload(pytom_tm)
            # check if plotting is indeed disabled
            # (reload is needed to prevent python import caching)
            self.assertFalse(reload(pytom_tm.extract).plotting_available)
            # assert that importing the plotting module fails completely
            with self.assertRaisesRegex(RuntimeError, "matplotlib and seaborn"):
                reload(pytom_tm.plotting)



import pathlib
import mrcfile
import argparse
import logging
import numpy.typing as npt
import numpy as np
import starfile
from contextlib import contextmanager
from operator import attrgetter


class ParseLogging(argparse.Action):
    """argparse.Action subclass to parse logging parameter from input scripts. Users can
    set these to info/debug."""

    def __call__(
        self, parser, namespace, values: str, option_string: str | None = None
    ):
        if values.upper() not in ["INFO", "DEBUG"]:
            parser.error(
                f"{option_string} log got an invalid option, "
                "set either to `info` or `debug` "
            )
        else:
            numeric_level = getattr(logging, values.upper(), None)
            setattr(namespace, self.dest, numeric_level)


class CheckDirExists(argparse.Action):
    """argparse.Action subclass to check if an expected input directory exists."""

    def __call__(
        self,
        parser,
        namespace,
        values: pathlib.Path,
        option_string: str | None = None,
    ):
        if not values.is_dir():
            parser.error(f"{option_string} got a file path that does not exist ")

        setattr(namespace, self.dest, values)


class CheckFileExists(argparse.Action):
    """argparse.Action subclass to check if an expected input file exists."""

    def __call__(
        self,
        parser,
        namespace,
        values: pathlib.Path,
        option_string: str | None = None,
    ):
        if not values.exists():
            parser.error(f"{option_string} got a file path that does not exist ")

        setattr(namespace, self.dest, values)


class LargerThanZero(argparse.Action):
    """argparse.Action subclass to constrain an input value to larger than zero only."""

    def __call__(
        self,
        parser,
        namespace,
        values: int | float,
        option_string: str | None = None,
    ):
        if values <= 0.0:
            parser.error(f"{option_string} must be larger than 0")

        setattr(namespace, self.dest, values)


class BetweenZeroAndOne(argparse.Action):
    """argparse.Action subclass to constrain an input value to a fraction, i.e. between
    0 and 1."""

    def __call__(
        self, parser, namespace, values: float, option_string: str | None = None
    ):
        if 1.0 <= values <= 0.0:
            parser.error(
                f"{option_string} is a fraction and can only range between 0 and 1"
            )

        setattr(namespace, self.dest, values)


class ParseSearch(argparse.Action):
    """argparse.Action subclass to restrict the search area of tomogram to these indices
    along an axis. Checks that these value are larger than zero and that the second
    value is larger than the first."""

    def __call__(
        self,
        parser,
        namespace,
        values: list[int, int],
        option_string: str | None = None,
    ):
        if not (0 <= values[0] < values[1]):
            parser.error(
                f"{option_string} start and end indices must be larger than 0 and end "
                "must be larger than start"
            )

        setattr(namespace, self.dest, values)


class ParseTiltAngles(argparse.Action):
    """argparse.Action subclass to parse tilt_angle info. The input can either be two
    floats that specify the tilt range for a continous wedge model. Alternatively can be
    a .tlt/.rawtlt file that specifies all the the tilt angles of the tilt-series to use
    for more refined wedge models."""

    def __call__(
        self,
        parser,
        namespace,
        values: list[str, str] | str,
        option_string: str | None = None,
    ):
        if len(values) == 2:  # two wedge angles provided the min and max
            try:
                values = sorted(list(map(float, values)))  # make them floats
                setattr(namespace, self.dest, values)
            except ValueError:
                parser.error(
                    f"{option_string} the two arguments provided could not be parsed "
                    "to floats"
                )
        elif len(values) == 1:
            values = pathlib.Path(values[0])
            if not values.exists() or values.suffix not in [".tlt", ".rawtlt"]:
                parser.error(
                    f"{option_string} provided tilt angle file does not exist or does "
                    "not have the right format"
                )
            setattr(namespace, self.dest, read_tlt_file(values))
        else:
            parser.error(f"{option_string} can only take one or two arguments")


class ParseGPUIndices(argparse.Action):
    """argparse.Action subclass to parse gpu indices. The input can either be an int or
    a list of ints that specify the gpu indices to be used.
    """

    def __call__(
        self,
        parser,
        namespace,
        values: list[int, ...],
        option_string: str | None = None,
    ):
        import cupy

        max_value = cupy.cuda.runtime.getDeviceCount()
        for val in values:
            if val < 0 or val >= max_value:
                parser.error(
                    f"{option_string} all gpu indices should be between 0 "
                    "and {max_value-1}"
                )

        setattr(namespace, self.dest, values)


class ParseDoseFile(argparse.Action):
    """argparse.Action subclass to parse a txt file contain information on accumulated
    dose per tilt."""

    def __call__(
        self, parser, namespace, values: str, option_string: str | None = None
    ):
        file_path = pathlib.Path(values)
        if not file_path.exists():
            parser.error(
                f"{option_string} provided dose accumulation file does not exist"
            )
        allowed_suffixes = [".txt"]
        if file_path.suffix not in allowed_suffixes:
            parser.error(
                f"{option_string}  provided dose accumulation file does not have the "
                f"right suffix, allowed are: {', '.join(allowed_suffixes)}"
            )
        setattr(namespace, self.dest, read_dose_file(file_path))


class ParseDefocus(argparse.Action):
    """argparse.Action subclass to read a defocus file, either from IMOD which adheres
    to their file format, or a txt file containing per line the defocus of each tilt."""

    def __call__(
        self, parser, namespace, values: str, option_string: str | None = None
    ):
        if values.endswith((".defocus", ".txt")):
            file_path = pathlib.Path(values)
            if not file_path.exists():
                parser.error(f"{option_string} provided defocus file does not exist")
            # in case of a file the parser attribute becomes a list of defocii
            setattr(namespace, self.dest, read_defocus_file(file_path))
        else:
            try:
                defocus = float(values)
            except ValueError:
                parser.error(f"{option_string} not possible to read defocus as float")
            # pass back as list so defocus always returns a list
            setattr(namespace, self.dest, [defocus])


class UnequalSpacingError(Exception):
    """Exception for an mrc file that has unequal spacing along the xyz dimensions
    annotated in its voxel size metadata."""

    pass


def write_angle_list(
    data: npt.NDArray[float],
    file_name: pathlib.Path,
    order: tuple[int, int, int] = (0, 2, 1),
):
    """Helper function to write angular search list from old PyTom to current module.
    Order had to be changed as old PyTom always stored it as Z1, Z2, X, and here it's
    Z1, X, Z2.

    @todo remove function
    """
    with open(file_name, "w") as fstream:
        for i in range(data.shape[1]):
            fstream.write(
                " ".join([str(x) for x in [data[j, i] for j in order]]) + "\n"
            )


@contextmanager
def _wrap_mrcfile_readers(func, *args, **kwargs):
    """Try to autorecover broken mrcfiles, assumes 'permissive' is a kwarg and not an
    arg"""
    try:
        mrc = func(*args, **kwargs)
    except ValueError as err:
        # see if permissive can safe this
        logging.debug(f"mrcfile raised the following error: {err}, will try to recover")
        kwargs["permissive"] = True
        mrc = func(*args, **kwargs)
        if mrc.data is not None:
            logging.warning(
                f"Loading {args[0]} in strict mode gave an error. "
                "However, loading with 'permissive=True' did generate data, make sure "
                "this is correct!"
            )
        else:
            logging.debug("Could not reasonably recover")
            raise ValueError(
                f"{args[0]} header or data is too corrupt to recover, please fix the "
                "header or data"
            ) from err
    yield mrc
    # this should only be called after the context exists
    mrc.close()


def read_mrc_meta_data(file_name: pathlib.Path) -> dict:
    """Read the metadata of provided MRC file path (using mrcfile) and return as dict.

    If the voxel size along the x, y, and z dimensions differs a lot (not within 3
    decimals) the function will raise an UnequalSpacingError as it could mean template
    matching on these volumes might not be consistent.

    Parameters
    ----------
    file_name: pathlib.Path
        path to an MRC file

    Returns
    -------
    metadata: dict
        a dictionary of the mrc metadata with key 'shape' containing the x,y,z
        dimensions of the file and key 'voxel_size' containing the voxel size along
        x, y, and z and dimensions in Å units
    """
    meta_data = {}
    with _wrap_mrcfile_readers(mrcfile.mmap, file_name) as mrc:
        meta_data["shape"] = tuple(map(int, attrgetter("nx", "ny", "nz")(mrc.header)))
        # allow small numerical inconsistencies in voxel size of MRC headers, sometimes
        # seen in Warp
        if not all(
            [
                np.round(mrc.voxel_size.x, 3) == np.round(s, 3)
                for s in attrgetter("y", "z")(mrc.voxel_size)
            ]
        ):
            raise UnequalSpacingError(
                "Input volume voxel spacing is not identical in each dimension!"
            )
        else:
            if not all(
                [mrc.voxel_size.x == s for s in attrgetter("y", "z")(mrc.voxel_size)]
            ):
                logging.warning(
                    "Voxel size annotation in MRC is slightly different between "
                    f"dimensions, namely {mrc.voxel_size}. It might be a tiny "
                    "numerical inaccuracy, but please ensure this is not problematic."
                )
            meta_data["voxel_size"] = float(mrc.voxel_size.x)
    return meta_data


def write_mrc(
    file_name: pathlib.Path,
    data: npt.NDArray[float],
    voxel_size: float,
    overwrite: bool = True,
    transpose: bool = True,
) -> None:
    """Write data to an MRC file. Data is transposed before writing as pytom internally
    uses xyz ordering and MRCs use zyx.

    Parameters
    ----------
    file_name: pathlib.Path
        path on disk to write the file to
    data: npt.NDArray[float]
        numpy array to write as MRC
    voxel_size: float
        voxel size of array to annotate in MRC header
    overwrite: bool, default True
        True (default) will overwrite current MRC on path, setting to False will error
        when writing to existing file
    transpose: bool, default True
        True (default) transpose array before writing, setting to False prevents this

    Returns
    -------
    """
    if data.dtype not in [np.float32, np.float16]:
        logging.warning(
            "data for mrc writing is not np.float32 or np.float16, will convert to "
            "np.float32"
        )
        data = data.astype(np.float32)
    mrcfile.write(
        file_name,
        data.T if transpose else data,
        voxel_size=voxel_size,
        overwrite=overwrite,
    )


def read_mrc(file_name: pathlib.Path, transpose: bool = True) -> npt.NDArray[float]:
    """Read an MRC file from disk. Data is transposed after reading as pytom internally
    uses xyz ordering and MRCs use zyx.

    Parameters
    ----------
    file_name: pathlib.Path
        path to file on disk
    transpose: bool, default True
        True (default) transposes the volume after reading, setting to False prevents
        transpose but probably not a good idea when using the functions from this module

    Returns
    -------
    data: npt.NDArray[float]
        returns the MRC data as a numpy array
    """
    with _wrap_mrcfile_readers(mrcfile.open, file_name) as mrc:
        data = np.ascontiguousarray(mrc.data.T) if transpose else mrc.data
    return data


def read_txt_file(file_name: pathlib.Path) -> list[float, ...]:
    """Read a txt file from disk with on each line a single float value.

    Parameters
    ----------
    file_name: pathlib.Path
        file on disk to read

    Returns
    -------
    output: list[float, ...]
        list of floats
    """
    with open(file_name) as fstream:
        lines = fstream.readlines()
    return list(map(float, [x.strip() for x in lines if not x.isspace()]))


def read_tlt_file(file_name: pathlib.Path) -> list[float, ...]:
    """Read a txt file from disk using read_txt_file(). File is expected to have tilt
    angles in degrees.

    Parameters
    ----------
    file_name: pathlib.Path
        file on disk to read

    Returns
    -------
    output: list[float, ...]
        list of floats with tilt angles
    """
    return read_txt_file(file_name)


def read_dose_file(file_name: pathlib.Path) -> list[float, ...]:
    """Read a txt file from disk using read_txt_file(). File is expected to have dose
    accumulation in e-/(Å^2).

    Parameters
    ----------
    file_name: pathlib.Path
        file on disk to read

    Returns
    -------
    output: list[float, ...]
        list of floats with accumulated dose
    """
    return read_txt_file(file_name)


def read_imod_defocus_file(file_name: pathlib.Path) -> list[float, ...]:
    """Read an IMOD style defocus file. This function can read version 2 and 3 defocus
    files. For format specification see:
    https://bio3d.colorado.edu/imod/doc/man/ctfphaseflip.html
    (section: Defocus File Format).

    Parameters
    ----------
    file_name: pathlib.Path
        file on disk to read

    Returns
    -------
    output: list[float, ...]
        list of floats with defocus (in μm)
    """
    with open(file_name) as fstream:
        lines = fstream.readlines()
    imod_defocus_version = float(lines[0].strip().split()[5])
    # imod defocus files have the values specified in nm:
    if imod_defocus_version == 2:  # file with one defocus value; data starts on line 0
        return [float(x.strip().split()[4]) * 1e-3 for x in lines]
    elif (
        imod_defocus_version == 3
    ):  # file with astigmatism; line 0 contains metadata that we do not need
        return [
            (float(x.strip().split()[4]) + float(x.strip().split()[5])) / 2 * 1e-3
            for x in lines[1:]
        ]
    else:
        raise ValueError("Invalid IMOD defocus file inversion, can only be 2 or 3.")


def read_defocus_file(file_name: pathlib.Path) -> list[float, ...]:
    """Read a defocus file with values in nm. Output returns defocus in μm.

    Depending on file suffix the function calls:
     - read_imod_defocus_file() for .defocus suffix
     - read_txt_file for .txt suffix

    Parameters
    ----------
    file_name: pathlib.Path
        file on disk to read

    Returns
    -------
    output: list[float, ...]
        list of floats with defocus (in μm)
    """
    if file_name.suffix == ".defocus":
        return read_imod_defocus_file(file_name)
    elif file_name.suffix == ".txt":
        return read_txt_file(file_name)
    else:
        raise ValueError("Defocus file needs to have format .defocus or .txt")


def parse_relion5_star_data(
    tomograms_star_path: pathlib.Path,
    tomogram_path: pathlib.Path,
    phase_flip_correction: bool = False,
    phase_shift: float = 0.0,
) -> tuple[float, list[float, ...], list[float, ...], list[dict, ...], int]:
    """Read RELION5 metadata from a project directory.

    Parameters
    ----------
    tomograms_star_path: pathlib.Path
        the tomograms.star from a RELION5 reconstruct job contains invariable metadata
        and points to a tilt series star file with fitted values
    tomogram_path: pathlib.Path
        path to the tomogram for template matching; we use the name to pattern match in
        the RELION5 star file
    phase_flip_correction: bool, default False
    phase_shift: float, default 0.0

    Returns
    -------
    tomogram_voxel_size, tilt_angles, dose_accumulation, ctf_params, defocus_handedness:
        tuple[float, list[float, ...], list[float, ...], list[dict, ...], int]
    """
    tomogram_id = tomogram_path.stem
    tomograms_star_data = starfile.read(tomograms_star_path)

    # match the tomo_id and check if viable
    matches = [
        i
        for i, x in enumerate(tomograms_star_data["rlnTomoName"])
        if tomogram_id.endswith(x)
    ]
    if len(matches) == 1:
        tomogram_meta_data = tomograms_star_data.loc[matches[0]]
    else:
        raise ValueError(
            f"{'Multiple' if len(matches) > 1 else 'Zero'} matches "
            f"of tomogram id: {tomogram_id}, "
            f"in RELION5 STAR file: {tomograms_star_path}. "
            "Aborting..."
        )

    # grab the path to tilt series star for tilt angles, defocus and dose
    tilt_series_star_path = pathlib.Path(
        tomogram_meta_data["rlnTomoTiltSeriesStarFile"]
    )
    # update the path to a location we can actually find from CD
    tilt_series_star_path = tomograms_star_path.parent.joinpath("tilt_series").joinpath(
        tilt_series_star_path.name
    )
    tilt_series_star_data = starfile.read(tilt_series_star_path)

    # we extract tilt angles, dose accumulation and ctf params
    # TODO We need to have an internal structure for tilt series meta data
    tilt_angles = list(tilt_series_star_data["rlnTomoNominalStageTiltAngle"])
    dose_accumulation = list(tilt_series_star_data["rlnMicrographPreExposure"])

    tomogram_voxel_size = float(
        tomogram_meta_data["rlnTomoTiltSeriesPixelSize"]
        * tomogram_meta_data["rlnTomoTomogramBinning"]
    )
    defocus_handedness = int(tomogram_meta_data["rlnTomoHand"])

    ctf_params = [
        {
            "defocus": defocus * 1e-10,
            "amplitude_contrast": tomogram_meta_data["rlnAmplitudeContrast"],
            "voltage": tomogram_meta_data["rlnVoltage"] * 1e3,
            "spherical_aberration": tomogram_meta_data["rlnSphericalAberration"] * 1e-3,
            "flip_phase": phase_flip_correction,
            "phase_shift_deg": phase_shift,  # RELION5 does not seem to store this
        }
        for defocus in (
            tilt_series_star_data.rlnDefocusV + tilt_series_star_data.rlnDefocusU
        )
        / 2
    ]

    return (
        tomogram_voxel_size,
        tilt_angles,
        dose_accumulation,
        ctf_params,
        defocus_handedness,
    )



import unittest
import pathlib
import numpy as np
import cupy as cp
import logging
from shutil import which
from contextlib import redirect_stdout, redirect_stderr
from io import StringIO
from pytom_tm import entry_points
from pytom_tm import io

# (command line function, function in entry_points file)
ENTRY_POINTS_TO_TEST = [
    ("pytom_create_mask.py", "pytom_create_mask"),
    ("pytom_create_template.py", "pytom_create_template"),
    ("pytom_match_template.py", "match_template"),
    ("pytom_extract_candidates.py", "extract_candidates"),
    ("pytom_merge_stars.py", "merge_stars"),
]
# Test if optional dependencies are installed
try:
    from pytom_tm import plotting  # noqa: F401
except RuntimeError:
    pass
else:
    ENTRY_POINTS_TO_TEST.append(("pytom_estimate_roc.py", "estimate_roc"))

# Input files for command line match_template
TEST_DATA = pathlib.Path(__file__).parent.joinpath("test_data")
TEMPLATE = TEST_DATA.joinpath("template.mrc")
MASK = TEST_DATA.joinpath("mask.mrc")
TOMOGRAM = TEST_DATA.joinpath("tomogram.mrc")
DESTINATION = TEST_DATA.joinpath("output")
TILT_ANGLES = TEST_DATA.joinpath("angles.rawtlt")
DOSE = TEST_DATA.joinpath("test_dose.txt")
DEFOCUS = TEST_DATA.joinpath("defocus.txt")
DEFOCUS_IMOD = (
    pathlib.Path(__file__).parent.joinpath("Data").joinpath("test_imod.defocus")
)
RELION5_TOMOGRAMS_STAR = pathlib.Path(__file__).parent.joinpath(
    "Data/relion5_project_example/Tomograms/job009/tomograms.star"
)
RELION5_TOMOGRAM = TEST_DATA.joinpath("rec_tomo200528_107.mrc")

# Initial logging level
LOG_LEVEL = logging.getLogger().level


def prep_argv(arg_dict):
    argv = []
    [
        argv.extend([k] + v.split()) if v != "" else argv.append(k)
        for k, v in arg_dict.items()
    ]
    return argv


class TestEntryPoints(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        DESTINATION.mkdir(parents=True)
        io.write_mrc(TEMPLATE, np.zeros((5, 5, 5), dtype=np.float32), 1)
        io.write_mrc(MASK, np.zeros((5, 5, 5), dtype=np.float32), 1)
        io.write_mrc(TOMOGRAM, np.zeros((10, 10, 10), dtype=np.float32), 1)
        io.write_mrc(RELION5_TOMOGRAM, np.zeros((10, 10, 10), dtype=np.float32), 1)
        np.savetxt(TILT_ANGLES, np.linspace(-50, 50, 35))
        np.savetxt(DOSE, np.linspace(0, 100, 35))
        np.savetxt(DEFOCUS, np.ones(35) * 3000)

    @classmethod
    def tearDownClass(cls) -> None:
        TEMPLATE.unlink()
        MASK.unlink()
        TOMOGRAM.unlink()
        RELION5_TOMOGRAM.unlink()
        TILT_ANGLES.unlink()
        DOSE.unlink()
        DEFOCUS.unlink()
        for f in DESTINATION.iterdir():
            f.unlink()  # should test specific output?
        DESTINATION.rmdir()
        TEST_DATA.rmdir()

    def test_entry_points_exist(self):
        for cli, fname in ENTRY_POINTS_TO_TEST:
            # test the command line function can be found
            self.assertIsNotNone(which(cli))
            # assert the entry_point be called with -h and exit cleanly
            # catch stdout to prevent shell polution
            func = getattr(entry_points, fname)
            dump = StringIO()
            with self.assertRaises(SystemExit) as ex, redirect_stdout(dump):
                func([cli, "-h"])
            dump.close()
            # check if the system return code is 0 (success)
            self.assertEqual(ex.exception.code, 0)

    def test_match_template(self):
        defaults = {
            "-t": str(TEMPLATE),
            "-m": str(MASK),
            "-v": str(TOMOGRAM),
            "-d": str(DESTINATION),
            "--angular-search": "35",
            "--tilt-angles": str(TILT_ANGLES),
            "--per-tilt-weighting": "",
            "--dose-accumulation": str(DOSE),
            "--defocus": str(DEFOCUS_IMOD),
            "--amplitude-contrast": "0.08",
            "--spherical-aberration": "2.7",
            "--voltage": "300",
            "--tomogram-ctf-model": "phase-flip",
            "-g": "0",
        }

        def start(arg_dict):  # simplify run
            entry_points.match_template(prep_argv(arg_dict))

        # test valid defocus arguments
        for z in [str(DEFOCUS_IMOD), str(DEFOCUS), "3000"]:
            arguments = defaults.copy()
            arguments["--defocus"] = z
            start(arguments)

        # test faulty args
        for z in ["asdf.txt", "asdf"]:
            dump = StringIO()
            with (
                self.assertRaises(SystemExit) as ex,
                redirect_stdout(dump),
                redirect_stderr(dump),
            ):
                arguments = defaults.copy()
                arguments["--defocus"] = z
                start(arguments)
            dump.close()
            # check if the system return code is 0 (success)
            self.assertEqual(ex.exception.code, 2)

        # remove per-tilt-weighting
        arguments = defaults.copy()
        arguments.pop("--per-tilt-weighting")
        start(arguments)

        with self.assertRaises(
            ValueError, msg="Missing CTF params should produce " "error"
        ):
            arguments = defaults.copy()
            arguments.pop("--voltage")
            start(arguments)

        # test angular search and particle diameter options
        with self.assertRaises(
            ValueError, msg="Missing angular search should raise " "an error."
        ):
            arguments = defaults.copy()
            arguments.pop("--angular-search")
            start(arguments)

        arguments = defaults.copy()
        arguments.pop("--angular-search")
        arguments["--particle-diameter"] = "50"
        # set low-pass to tune the search to lower degree
        arguments["--low-pass"] = "50"
        start(arguments)

        # phase randomization test
        arguments = defaults.copy()
        arguments["-r"] = ""
        start(arguments)
        # test if we can set the rng seed, see issue #194
        arguments["--rng-seed"] = "42"
        start(arguments)

        # test debug files
        arguments = defaults.copy()
        arguments["--log"] = "debug"
        start(arguments)
        # these files will only exist if the test managed to set the logging correctly
        self.assertTrue(
            DESTINATION.joinpath("template_psf.mrc").exists(),
            msg="File should exist in debug mode",
        )
        self.assertTrue(
            DESTINATION.joinpath("template_convolved.mrc").exists(),
            msg="File should exist in debug mode",
        )

        # reset the log level after the entry point modified it
        logging.basicConfig(level=LOG_LEVEL, force=True)

        # test providing invalid gpu indices
        n_devices = cp.cuda.runtime.getDeviceCount()
        for indices in ["-1", f"0 {n_devices}"]:
            dump = StringIO()
            with (
                self.assertRaises(SystemExit) as ex,
                redirect_stdout(dump),
                redirect_stderr(dump),
            ):
                arguments = defaults.copy()
                arguments["-g"] = indices
                start(arguments)
            self.assertIn("gpu indices", dump.getvalue())
            dump.close()

        # test relion5 metadata reading
        arguments = defaults.copy()
        [
            arguments.pop(x)
            for x in [
                "--tilt-angles",
                "--per-tilt-weighting",
                "--dose-accumulation",
                "--defocus",
                "--amplitude-contrast",
                "--spherical-aberration",
                "--voltage",
            ]
        ]
        arguments["-v"] = str(RELION5_TOMOGRAM)
        arguments["--relion5-tomograms-star"] = str(RELION5_TOMOGRAMS_STAR)
        start(arguments)

        with self.assertRaises(
            ValueError, msg="Missing tilt angles should raise an error."
        ):
            arguments.pop("--relion5-tomograms-star")
            start(arguments)



import unittest
import pathlib
import numpy as np
import voltools as vt
import mrcfile
from tempfile import TemporaryDirectory
from pytom_tm.mask import spherical_mask
from pytom_tm.angles import angle_to_angle_list
from pytom_tm.tmjob import TMJob, TMJobError, load_json_to_tmjob, get_defocus_offsets
from pytom_tm.io import read_mrc, write_mrc, UnequalSpacingError
from pytom_tm.extract import extract_particles
from testing_utils import CTF_PARAMS, ACCUMULATED_DOSE, TILT_ANGLES


TOMO_SHAPE = (100, 107, 59)
TEMPLATE_SIZE = 13
LOCATION = (77, 26, 40)
ANGLE_ID = 100
ANGULAR_SEARCH = "38.53"
TEMP_DIR = TemporaryDirectory()
TEST_DATA_DIR = pathlib.Path(TEMP_DIR.name)
TEST_TOMOGRAM = TEST_DATA_DIR.joinpath("tomogram.mrc")
TEST_BROKEN_TOMOGRAM_MASK = TEST_DATA_DIR.joinpath("broken_tomogram_mask.mrc")
TEST_WRONG_SIZE_TOMO_MASK = TEST_DATA_DIR.joinpath("wrong_size_tomogram_mask.mrc")
TEST_EXTRACTION_MASK_OUTSIDE = TEST_DATA_DIR.joinpath("extraction_mask_outside.mrc")
TEST_EXTRACTION_MASK_INSIDE = TEST_DATA_DIR.joinpath("extraction_mask_inside.mrc")
TEST_TEMPLATE = TEST_DATA_DIR.joinpath("template.mrc")
TEST_TEMPLATE_UNEQUAL_SPACING = TEST_DATA_DIR.joinpath("template_unequal_spacing.mrc")
TEST_TEMPLATE_WRONG_VOXEL_SIZE = TEST_DATA_DIR.joinpath("template_voxel_error_test.mrc")
TEST_MASK = TEST_DATA_DIR.joinpath("mask.mrc")
TEST_SCORES = TEST_DATA_DIR.joinpath("tomogram_scores.mrc")
TEST_ANGLES = TEST_DATA_DIR.joinpath("tomogram_angles.mrc")
TEST_CUSTOM_ANGULAR_SEARCH = TEST_DATA_DIR.joinpath("custom_angular_search.txt")
TEST_WHITENING_FILTER = TEST_DATA_DIR.joinpath("tomogram_whitening_filter.npy")
TEST_JOB_JSON = TEST_DATA_DIR.joinpath("tomogram_job.json")
TEST_JOB_JSON_WHITENING = TEST_DATA_DIR.joinpath("tomogram_job_whitening.json")
TEST_JOB_OLD_VERSION = TEST_DATA_DIR.joinpath("tomogram_job_old_version.json")


class TestTMJob(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Create template, mask and tomogram
        volume = np.zeros(TOMO_SHAPE, dtype=np.float32)
        template = np.zeros((TEMPLATE_SIZE,) * 3, dtype=np.float32)
        template[3:8, 4:8, 3:7] = 1.0
        template[7, 8, 5:7] = 1.0
        mask = spherical_mask(TEMPLATE_SIZE, 5, 0.5)
        rotation = angle_to_angle_list(float(ANGULAR_SEARCH))[ANGLE_ID]

        volume[
            LOCATION[0] - TEMPLATE_SIZE // 2 : LOCATION[0]
            + TEMPLATE_SIZE // 2
            + TEMPLATE_SIZE % 2,
            LOCATION[1] - TEMPLATE_SIZE // 2 : LOCATION[1]
            + TEMPLATE_SIZE // 2
            + TEMPLATE_SIZE % 2,
            LOCATION[2] - TEMPLATE_SIZE // 2 : LOCATION[2]
            + TEMPLATE_SIZE // 2
            + TEMPLATE_SIZE % 2,
        ] = vt.transform(
            template,
            rotation=rotation,
            rotation_units="rad",
            rotation_order="rzxz",
            device="cpu",
        )

        # add some noise
        rng = np.random.default_rng(0)
        volume += rng.normal(loc=0, scale=0.1, size=volume.shape)

        # extraction mask
        extraction_mask_outside = np.zeros(TOMO_SHAPE, dtype=np.float32)
        extraction_mask_outside[20:40, 60:80, 10:30] = 1
        extraction_mask_inside = np.zeros(TOMO_SHAPE, dtype=np.float32)
        extraction_mask_inside[70:90, 15:35, 30:50] = 1

        TEST_DATA_DIR.mkdir(exist_ok=True)
        write_mrc(TEST_EXTRACTION_MASK_OUTSIDE, extraction_mask_outside, 1.0)
        write_mrc(TEST_EXTRACTION_MASK_INSIDE, extraction_mask_inside, 1.0)
        write_mrc(TEST_MASK, mask, 1.0)
        write_mrc(TEST_TEMPLATE, template, 1.0)
        write_mrc(TEST_TEMPLATE_WRONG_VOXEL_SIZE, template, 1.5)
        mrcfile.write(
            TEST_TEMPLATE_UNEQUAL_SPACING,
            template,
            voxel_size=(1.5, 1.0, 2.0),
            overwrite=True,
        )
        write_mrc(TEST_TOMOGRAM, volume, 1.0)

        # do a run without splitting to compare against
        job = TMJob(
            "0",
            10,
            TEST_TOMOGRAM,
            TEST_TEMPLATE,
            TEST_MASK,
            TEST_DATA_DIR,
            angle_increment=ANGULAR_SEARCH,
            voxel_size=1.0,
        )
        score, angle = job.start_job(0, return_volumes=True)
        write_mrc(TEST_SCORES, score, job.voxel_size)
        write_mrc(TEST_ANGLES, angle, job.voxel_size)
        job.write_to_json(TEST_JOB_JSON)

        np.savetxt(TEST_CUSTOM_ANGULAR_SEARCH, np.random.rand(100, 3))

        # create job with spectrum whitening
        job = TMJob(
            "0",
            10,
            TEST_TOMOGRAM,
            TEST_TEMPLATE,
            TEST_MASK,
            TEST_DATA_DIR,
            angle_increment=90.00,
            voxel_size=1.0,
            whiten_spectrum=True,
        )
        job.write_to_json(TEST_JOB_JSON_WHITENING)

        # write broken tomogram mask
        broken_tomogram_mask = np.zeros(TOMO_SHAPE, dtype=np.float32)
        write_mrc(TEST_BROKEN_TOMOGRAM_MASK, broken_tomogram_mask, 1.0)

        # write wrong size tomogram mask
        size = list(TOMO_SHAPE)
        size[0] += 1
        wrong_size_tomogram_mask = np.ones(tuple(size), dtype=np.float32)
        write_mrc(TEST_WRONG_SIZE_TOMO_MASK, wrong_size_tomogram_mask, 1.0)

    @classmethod
    def tearDownClass(cls) -> None:
        TEMP_DIR.cleanup()

    def setUp(self):
        self.job = TMJob(
            "0",
            10,
            TEST_TOMOGRAM,
            TEST_TEMPLATE,
            TEST_MASK,
            TEST_DATA_DIR,
            angle_increment=ANGULAR_SEARCH,
            voxel_size=1.0,
        )

    def test_tm_job_errors(self):
        with self.assertRaises(
            ValueError,
            msg="Different voxel size in tomogram and template and no voxel size "
            "provided should raise a ValueError",
        ):
            TMJob(
                "0",
                10,
                TEST_TOMOGRAM,
                TEST_TEMPLATE_WRONG_VOXEL_SIZE,
                TEST_MASK,
                TEST_DATA_DIR,
            )

        with self.assertRaises(
            UnequalSpacingError, msg="Unequal spacing should raise specific Error"
        ):
            TMJob(
                "0",
                10,
                TEST_TOMOGRAM,
                TEST_TEMPLATE_UNEQUAL_SPACING,
                TEST_MASK,
                TEST_DATA_DIR,
            )

        # test searches raise correct errors
        for param in ["search_x", "search_y", "search_z"]:
            with self.assertRaises(
                ValueError, msg="Invalid start index in search should raise ValueError"
            ):
                TMJob(
                    "0",
                    10,
                    TEST_TOMOGRAM,
                    TEST_TEMPLATE,
                    TEST_MASK,
                    TEST_DATA_DIR,
                    voxel_size=1.0,
                    **{param: [-10, 100]},
                )
            with self.assertRaises(
                ValueError, msg="Invalid start index in search should raise ValueError"
            ):
                TMJob(
                    "0",
                    10,
                    TEST_TOMOGRAM,
                    TEST_TEMPLATE,
                    TEST_MASK,
                    TEST_DATA_DIR,
                    voxel_size=1.0,
                    **{param: [110, 130]},
                )
            with self.assertRaises(
                ValueError, msg="Invalid end index in search should raise ValueError"
            ):
                TMJob(
                    "0",
                    10,
                    TEST_TOMOGRAM,
                    TEST_TEMPLATE,
                    TEST_MASK,
                    TEST_DATA_DIR,
                    voxel_size=1.0,
                    **{param: [0, 120]},
                )
        # Test broken angle input
        with self.assertRaisesRegex(TMJobError, "Invalid angular search"):
            TMJob(
                "0",
                10,
                TEST_TOMOGRAM,
                TEST_TEMPLATE,
                TEST_MASK,
                TEST_DATA_DIR,
                angle_increment="1.2.3",
                voxel_size=1.0,
            )

        # Test broken template mask
        with self.assertRaisesRegex(ValueError, str(TEST_BROKEN_TOMOGRAM_MASK)):
            TMJob(
                "0",
                10,
                TEST_TOMOGRAM,
                TEST_TEMPLATE,
                TEST_MASK,
                TEST_DATA_DIR,
                angle_increment=ANGULAR_SEARCH,
                voxel_size=1.0,
                tomogram_mask=TEST_BROKEN_TOMOGRAM_MASK,
            )
        # Test wrong size template mask
        with self.assertRaisesRegex(ValueError, str(TOMO_SHAPE)):
            TMJob(
                "0",
                10,
                TEST_TOMOGRAM,
                TEST_TEMPLATE,
                TEST_MASK,
                TEST_DATA_DIR,
                angle_increment=ANGULAR_SEARCH,
                voxel_size=1.0,
                tomogram_mask=TEST_WRONG_SIZE_TOMO_MASK,
            )

    def test_tm_job_copy(self):
        copy = self.job.copy()
        self.assertIsNot(
            self.job, copy, msg="Copying the job should create a new object."
        )
        self.assertEqual(
            TOMO_SHAPE,
            copy.tomo_shape,
            msg="Tomogram shape not correct in job, perhaps transpose issue?",
        )

    def test_tm_job_weighting_options(self):
        # run with all options
        job = TMJob(
            "0",
            10,
            TEST_TOMOGRAM,
            TEST_TEMPLATE,
            TEST_MASK,
            TEST_DATA_DIR,
            angle_increment=90.00,
            voxel_size=1.0,
            low_pass=10,
            high_pass=100,
            dose_accumulation=ACCUMULATED_DOSE,
            ctf_data=CTF_PARAMS,
            tilt_angles=TILT_ANGLES,
            whiten_spectrum=True,
            tilt_weighting=True,
            defocus_handedness=1,
        )
        score, angle = job.start_job(0, return_volumes=True)
        self.assertEqual(
            score.shape, job.tomo_shape, msg="TMJob with all options failed"
        )

        # run with only tilt weighting
        # (in test_weights different options are tested for create_wedge)
        job = TMJob(
            "0",
            10,
            TEST_TOMOGRAM,
            TEST_TEMPLATE,
            TEST_MASK,
            TEST_DATA_DIR,
            angle_increment=90.00,
            voxel_size=1.0,
            dose_accumulation=ACCUMULATED_DOSE,
            ctf_data=CTF_PARAMS,
            tilt_angles=TILT_ANGLES,
            tilt_weighting=True,
        )
        score, angle = job.start_job(0, return_volumes=True)
        self.assertEqual(
            score.shape, job.tomo_shape, msg="TMJob with only wedge creation failed"
        )

        # run with only bandpass (in test_weights bandpass option are tested)
        job = TMJob(
            "0",
            10,
            TEST_TOMOGRAM,
            TEST_TEMPLATE,
            TEST_MASK,
            TEST_DATA_DIR,
            angle_increment=90.00,
            voxel_size=1.0,
            low_pass=10,
            high_pass=100,
        )
        score, angle = job.start_job(0, return_volumes=True)
        self.assertEqual(
            score.shape, job.tomo_shape, msg="TMJob with only band-pass failed"
        )

        # run with only spectrum whitening
        # (in test_weights the whitening filter is tested)
        job = TMJob(
            "0",
            10,
            TEST_TOMOGRAM,
            TEST_TEMPLATE,
            TEST_MASK,
            TEST_DATA_DIR,
            angle_increment=90.00,
            voxel_size=1.0,
            whiten_spectrum=True,
        )
        score, angle = job.start_job(0, return_volumes=True)
        self.assertEqual(
            score.shape, job.tomo_shape, msg="TMJob with only whitening filter failed"
        )

        # load the whitening filter from previous job to compare against
        whitening_filter = np.load(TEST_WHITENING_FILTER)
        job = TMJob(
            "0",
            10,
            TEST_TOMOGRAM,
            TEST_TEMPLATE,
            TEST_MASK,
            TEST_DATA_DIR,
            angle_increment=90.00,
            voxel_size=1.0,
            whiten_spectrum=True,
            search_y=[10, 90],
        )
        new_whitening_filter = np.load(TEST_WHITENING_FILTER)
        self.assertNotEqual(
            whitening_filter.shape,
            new_whitening_filter.shape,
            msg="After reducing the search region along the largest dimension the "
            "whitening filter should have less sampling points",
        )
        self.assertEqual(
            new_whitening_filter.shape,
            (max(job.search_size) // 2 + 1,),
            msg="The whitening filter does not have the expected size, it should be "
            "equal (x // 2) + 1, where x is the largest dimension of the search box.",
        )

        # TMJob with none of these weighting options is tested in all other runs
        # in this file.

    def test_load_json_to_tmjob(self):
        # check base job loading
        job = load_json_to_tmjob(TEST_JOB_JSON)
        self.assertIsInstance(
            job, TMJob, msg="TMJob could not be properly loaded from disk."
        )

        # check job loading and preventing whitening filter recalculation
        with self.assertNoLogs(level="INFO"):
            _ = load_json_to_tmjob(TEST_JOB_JSON_WHITENING, load_for_extraction=True)
        with self.assertLogs(level="INFO") as cm:
            _ = load_json_to_tmjob(TEST_JOB_JSON_WHITENING, load_for_extraction=False)
        self.assertIn("Estimating whitening filter...", "".join(cm.output))

        # turn current job into 0.6.0 job with ctf params
        job.pytom_tm_version_number = "0.6.0"
        job.ctf_data = []
        for ctf in CTF_PARAMS:
            job.ctf_data.append(ctf.copy())
            del job.ctf_data[-1]["phase_shift_deg"]
        job.write_to_json(TEST_JOB_OLD_VERSION)

        # test backward compatibility with the update to 0.6.1
        job = load_json_to_tmjob(TEST_JOB_OLD_VERSION)
        self.assertEqual(job.ctf_data[0]["phase_shift_deg"], 0.0)

    def test_custom_angular_search(self):
        with TemporaryDirectory() as data_dir:
            data_dir = pathlib.Path(data_dir)
            job = TMJob(
                "0",
                10,
                TEST_TOMOGRAM,
                TEST_TEMPLATE,
                TEST_MASK,
                data_dir,
                angle_increment=TEST_CUSTOM_ANGULAR_SEARCH,
                voxel_size=1.0,
            )
            self.assertEqual(
                job.rotation_file,
                TEST_CUSTOM_ANGULAR_SEARCH,
                msg="Passing a custom angular search file to TMJob failed.",
            )

            # Also test extraction works with custom angle file

            scores, angles = job.start_job(0, return_volumes=True)
            write_mrc(data_dir / "tomogram_scores.mrc", scores, job.voxel_size)
            write_mrc(data_dir / "tomogram_angles.mrc", angles, job.voxel_size)
            df, scores = extract_particles(job, 5, 100, create_plot=False)
            self.assertNotEqual(
                len(scores), 0, msg="Here we expect to get some annotations."
            )

    def test_tm_job_split_volume(self):
        # Splitting the volume into smaller boxes than the template
        # should not raise an error
        _ = self.job.split_volume_search((10, 3, 2))
        # Reset
        self.job.sub_jobs = []
        # Make sure that asking for more splits than pixels results in
        # just a pixel number of jobs
        with self.assertWarnsRegex(RuntimeWarning, "More splits than pixels"):
            self.job.split_volume_search((TOMO_SHAPE[0] + 42, 1, 1))
        self.assertEqual(len(self.job.sub_jobs), TOMO_SHAPE[0])
        # Reset
        self.job.sub_jobs = []
        # Negative splits should fail
        with self.assertRaisesRegex(RuntimeError, "splits=-42"):
            self.job.split_volume_search((-42, 1, 1))
        sub_jobs = self.job.split_volume_search((2, 3, 2))
        stats = []
        for x in sub_jobs:
            stats.append(x.start_job(0))
            job_scores = TEST_DATA_DIR.joinpath(f"tomogram_scores_{x.job_key}.mrc")
            job_angles = TEST_DATA_DIR.joinpath(f"tomogram_angles_{x.job_key}.mrc")
            self.assertTrue(
                job_scores.exists(), msg="Expected output from job does not exist."
            )
            self.assertTrue(
                job_angles.exists(), msg="Expected output from job does not exist."
            )
        score, angle = self.job.merge_sub_jobs(stats)
        ind = np.unravel_index(score.argmax(), score.shape)

        self.assertTrue(score.max() > 0.931, msg="lcc max value lower than expected")
        self.assertEqual(ANGLE_ID, angle[ind])
        self.assertSequenceEqual(LOCATION, ind)

        # Small difference in the edge regions of the split dimension. This is because
        # the cross correlation function is not well defined in the boundary area, only
        # a small part of the template is correlated here (and we are not really
        # interested in it). Probably the inaccuracy in this area becomes more apparent
        # when splitting into subvolumes due to a smaller number of sampling points in
        # Fourier space.
        ok_region = slice(TEMPLATE_SIZE // 2, -TEMPLATE_SIZE // 2)
        score_diff = np.abs(
            score[ok_region, ok_region, ok_region]
            - read_mrc(TEST_SCORES)[ok_region, ok_region, ok_region]
        ).sum()

        self.assertAlmostEqual(
            score_diff, 0, places=1, msg="score diff should not be larger than 0.01"
        )
        # There is some race condition that sometimes gives
        # a different angle for 1 specific point.
        # This point is deterministic but different per machine
        # We suspect it has something to do with the FFT padding
        # See https://github.com/SBC-Utrecht/pytom-match-pick/pull/163

        # angle_diff = np.abs(
        #    angle[ok_region, ok_region, ok_region] -
        #    read_mrc(TEST_ANGLES)[ok_region, ok_region, ok_region]
        #    ).sum()

        # self.assertAlmostEqual(angle_diff, 0, places=1,
        #    msg='angle diff should not change')

        # get search statistics before and after splitting
        split_stats = self.job.job_stats
        reference_stats = load_json_to_tmjob(TEST_JOB_JSON).job_stats
        self.assertEqual(
            split_stats["search_space"],
            reference_stats["search_space"],
            msg="Search space should remain identical upon subvolume splitting.",
        )
        self.assertAlmostEqual(
            split_stats["std"],
            reference_stats["std"],
            places=3,
            msg="Standard deviation over template matching with subvolume splitting "
            "should be almost identical.",
        )

    def test_splitting_with_tomogram_mask(self):
        job = self.job.copy()
        job.tomogram_mask = TEST_EXTRACTION_MASK_INSIDE
        job.split_volume_search((10, 10, 10))
        self.assertLess(len(job.sub_jobs), 10 * 10 * 10)

    def test_splitting_with_offsets(self):
        # check if subjobs have correct offsets for the main job,
        # the last sub job will have the largest errors
        job = TMJob(
            "0",
            10,
            TEST_TOMOGRAM,
            TEST_TEMPLATE,
            TEST_MASK,
            TEST_DATA_DIR,
            angle_increment=ANGULAR_SEARCH,
            voxel_size=1.0,
            search_x=[9, 90],
            search_y=[25, 102],
            search_z=[19, 54],
        )
        # split along each dimension and get only the last sub job
        last_sub_job = job.split_volume_search((2, 3, 2))[-1]
        # Make sure the start of the data + size of the last subjob
        # result in the correct size
        final_size = [
            i + j for i, j in zip(last_sub_job.whole_start, last_sub_job.sub_step)
        ]
        self.assertEqual(
            final_size,
            job.search_size,
            msg="the last subjobs (unique) start position plus its size should equal "
            "the search size of the main job",
        )

    def test_tm_job_split_angles(self):
        sub_jobs = self.job.split_rotation_search(3)
        stats = []
        for x in sub_jobs:
            stats.append(x.start_job(0))
            job_scores = TEST_DATA_DIR.joinpath(f"tomogram_scores_{x.job_key}.mrc")
            job_angles = TEST_DATA_DIR.joinpath(f"tomogram_angles_{x.job_key}.mrc")
            self.assertTrue(
                job_scores.exists(), msg="Expected output from job does not exist."
            )
            self.assertTrue(
                job_angles.exists(), msg="Expected output from job does not exist."
            )
        score, angle = self.job.merge_sub_jobs(stats)
        ind = np.unravel_index(score.argmax(), score.shape)

        self.assertTrue(score.max() > 0.931, msg="lcc max value lower than expected")
        self.assertEqual(ANGLE_ID, angle[ind])
        self.assertSequenceEqual(LOCATION, ind)

        self.assertTrue(
            np.abs(score - read_mrc(TEST_SCORES)).sum() == 0,
            msg="split rotation search should be identical",
        )
        self.assertTrue(
            np.abs(angle - read_mrc(TEST_ANGLES)).sum() == 0,
            msg="split rotation search should be identical",
        )

        # get search statistics before and after splitting
        split_stats = self.job.job_stats
        reference_stats = load_json_to_tmjob(TEST_JOB_JSON).job_stats
        self.assertEqual(
            split_stats["search_space"],
            reference_stats["search_space"],
            msg="Search space should remain identical upon angular search splitting.",
        )
        self.assertAlmostEqual(
            split_stats["std"],
            reference_stats["std"],
            places=6,
            msg="Standard deviation of template matching with angular search split "
            "should be almost identical.",
        )

    def test_tm_job_half_precision(self):
        job = TMJob(
            "0",
            10,
            TEST_TOMOGRAM,
            TEST_TEMPLATE,
            TEST_MASK,
            TEST_DATA_DIR,
            angle_increment=ANGULAR_SEARCH,
            voxel_size=1.0,
            output_dtype=np.float16,
        )
        s, a = job.start_job(0, return_volumes=True)
        self.assertEqual(s.dtype, np.float16)
        self.assertEqual(a.dtype, np.float32)

    def test_extractions(self):
        self.job.tomo_id = "rec_" + self.job.tomo_id
        scores, angles = self.job.start_job(0, return_volumes=True)
        # set the appropriate headers when writing!
        write_mrc(
            TEST_DATA_DIR.joinpath(f"{self.job.tomo_id}_scores.mrc"),
            scores,
            self.job.voxel_size,
        )
        write_mrc(
            TEST_DATA_DIR.joinpath(f"{self.job.tomo_id}_angles.mrc"),
            angles,
            self.job.voxel_size,
        )

        # extract particles after running the job
        df, scores = extract_particles(self.job, 5, 100, create_plot=False)
        self.assertNotEqual(
            len(scores), 0, msg="Here we expect to get some annotations."
        )

        # extract particles in relion5 style
        df_rel5, scores = extract_particles(
            self.job, 5, 100, create_plot=False, relion5_compat=True
        )
        for column in (
            "rlnCenteredCoordinateXAngst",
            "rlnCenteredCoordinateYAngst",
            "rlnCenteredCoordinateZAngst",
            "rlnTomoName",
            "rlnTomoTiltSeriesPixelSize",
        ):
            self.assertTrue(
                column in df_rel5.columns,
                msg=f"Expected {column} in relion5 dataframe.",
            )
        centered_location = (
            np.array(LOCATION) - (np.array(TOMO_SHAPE) / 2 - 1)
        ) * self.job.voxel_size
        diff = np.abs(np.array(df_rel5.iloc[0, 0:3]) - centered_location).sum()
        self.assertEqual(
            diff,
            0,
            msg="relion5 compat mode should return a centered "
            "location of the object",
        )
        self.assertNotIn("rec_", df_rel5["rlnTomoName"][0])

        # test extraction mask that does not cover the particle
        df, scores = extract_particles(
            self.job,
            5,
            100,
            tomogram_mask_path=TEST_EXTRACTION_MASK_OUTSIDE,
            create_plot=False,
        )
        self.assertEqual(
            len(scores),
            0,
            msg="Length of returned list should be 0 after applying mask where the "
            "object is not in the region of interest.",
        )
        # test if the extraction mask can be grabbed from the job instead
        job = self.job.copy()
        job.tomogram_mask = TEST_EXTRACTION_MASK_OUTSIDE
        df, scores = extract_particles(
            job,
            5,
            100,
            create_plot=False,
        )
        self.assertEqual(
            len(scores),
            0,
            msg="Length of returned list should be 0 after applying mask where the "
            "object is not in the region of interest.",
        )
        # test if all masks are ignored if ignore_tomogram_mask=True
        # and that a warning is raised
        with self.assertLogs(level="WARNING") as cm:
            df, scores = extract_particles(
                job,
                5,
                100,
                tomogram_mask_path=TEST_EXTRACTION_MASK_OUTSIDE,
                create_plot=False,
                ignore_tomogram_mask=True,
            )
        # Test if expected warning is logged
        for o in cm.output:
            if "Ignoring tomogram mask" in o:
                break
        else:
            # break is not hit
            self.fail("expected warning is not logged")  # pragma: no cover
        self.assertNotEqual(
            len(scores),
            0,
            msg="We would expect some annotations if all tomogram masks are ignored",
        )

        # test mask that covers the particle
        # and should override the one now attached to the job
        df, scores = extract_particles(
            job,
            5,
            100,
            tomogram_mask_path=TEST_EXTRACTION_MASK_INSIDE,
            create_plot=False,
        )
        self.assertNotEqual(
            len(scores),
            0,
            msg="We expected a detected particle with a extraction mask that "
            "covers the object.",
        )

        # test mask that is the wrong size raises an error
        with self.assertRaisesRegex(ValueError, str(TOMO_SHAPE)):
            _, _ = extract_particles(
                job,
                5,
                100,
                tomogram_mask_path=TEST_WRONG_SIZE_TOMO_MASK,
                create_plot=False,
            )

        # Also test the raise if it somehow got attached to the job
        job = self.job.copy()
        job.tomogram_mask = TEST_WRONG_SIZE_TOMO_MASK
        with self.assertRaisesRegex(ValueError, str(TOMO_SHAPE)):
            _, _ = extract_particles(
                job,
                5,
                100,
                create_plot=False,
            )

        # Test exraction with tophat filter and plotting
        df, scores = extract_particles(
            job,
            5,
            100,
            tomogram_mask_path=TEST_EXTRACTION_MASK_INSIDE,
            create_plot=True,
            tophat_filter=True,
        )
        self.assertNotEqual(
            len(scores),
            0,
            msg="We expected a detected particle with a extraction mask that "
            "covers the object.",
        )
        # We don't look for the plots, they might be skipped if no plotting is available

    def test_get_defocus_offsets(self):
        tilt_angles = list(range(-51, 54, 3))
        x_offset_um = 200 * 13.79 * 1e-4
        z_offset_um = 100 * 13.79 * 1e-4
        defocus_offsets = get_defocus_offsets(x_offset_um, z_offset_um, tilt_angles)
        self.assertEqual(
            len(defocus_offsets),
            len(tilt_angles),
            msg="get_defocus_offsets did not return a list with the same length as "
            "the number of tilt_angles",
        )
        defocus_offsets_inverted = get_defocus_offsets(
            x_offset_um, z_offset_um, tilt_angles, invert_handedness=True
        )
        # only the offset at the 0 degrees tilt is expected to be identical,
        # so the test checks if exactly one element is the same
        self.assertTrue(
            (defocus_offsets == defocus_offsets_inverted).sum() == 1,
            msg="inverted handedness should have one identical offset",
        )



import unittest
import voltools as vt
import numpy as np
from pytom_tm.matching import TemplateMatchingGPU
from pytom_tm.mask import spherical_mask
from pytom_tm.angles import angle_to_angle_list


class TestTM(unittest.TestCase):
    def setUp(self):
        self.t_size = 12
        self.volume = np.zeros((100,) * 3, dtype=float)
        self.template = np.zeros((self.t_size,) * 3, dtype=float)
        self.template[3:8, 4:8, 3:7] = 1.0
        self.template[7, 8, 5:7] = 1.0
        self.mask = spherical_mask(self.t_size, 5, 0.5)
        self.gpu_id = "gpu:0"
        self.angles = angle_to_angle_list(38.53)

    def test_search_spherical_mask(self):
        angle_id = 100
        rotation = self.angles[angle_id]
        loc = (77, 26, 40)
        self.volume[
            loc[0] - self.t_size // 2 : loc[0] + self.t_size // 2,
            loc[1] - self.t_size // 2 : loc[1] + self.t_size // 2,
            loc[2] - self.t_size // 2 : loc[2] + self.t_size // 2,
        ] = vt.transform(
            self.template,
            rotation=rotation,
            rotation_units="rad",
            rotation_order="rzxz",
            device="cpu",
        )

        tm = TemplateMatchingGPU(
            0,
            0,
            self.volume,
            self.template,
            self.mask,
            self.angles,
            list(range(len(self.angles))),
        )
        score_volume, angle_volume, stats = tm.run()

        ind = np.unravel_index(score_volume.argmax(), self.volume.shape)
        self.assertTrue(
            score_volume.max() > 0.99, msg="lcc max value lower than expected"
        )
        self.assertEqual(angle_id, angle_volume[ind])
        self.assertSequenceEqual(loc, ind)
        expected_search_space = len(self.angles) * self.volume.size
        self.assertEqual(
            stats["search_space"],
            expected_search_space,
            msg="Search space should exactly equal this value",
        )
        self.assertAlmostEqual(
            stats["std"],
            0.005163,
            places=5,
            msg="Standard deviation of the search should be almost equal",
        )

    def test_search_non_spherical_mask(self):
        angle_id = 100
        rotation = self.angles[angle_id]
        loc = (77, 26, 40)
        self.volume[
            loc[0] - self.t_size // 2 : loc[0] + self.t_size // 2,
            loc[1] - self.t_size // 2 : loc[1] + self.t_size // 2,
            loc[2] - self.t_size // 2 : loc[2] + self.t_size // 2,
        ] = vt.transform(
            self.template,
            rotation=rotation,
            rotation_units="rad",
            rotation_order="rzxz",
            device="cpu",
        )

        tm = TemplateMatchingGPU(
            0,
            0,
            self.volume,
            self.template,
            self.mask,
            self.angles,
            list(range(len(self.angles))),
            mask_is_spherical=False,
        )
        score_volume, angle_volume, stats = tm.run()

        ind = np.unravel_index(score_volume.argmax(), self.volume.shape)
        self.assertTrue(
            score_volume.max() > 0.99, msg="lcc max value lower than expected"
        )
        self.assertEqual(angle_id, angle_volume[ind])
        self.assertSequenceEqual(loc, ind)

        expected_search_space = len(self.angles) * self.volume.size
        self.assertEqual(
            stats["search_space"],
            expected_search_space,
            msg="Search space should exactly equal this value",
        )
        self.assertAlmostEqual(
            stats["std"],
            0.005187,
            places=4,
            msg="Standard deviation of the search should be almost equal",
        )

    def test_search_noise_correction(self):
        angle_id = 100
        rotation = self.angles[angle_id]
        loc = (77, 26, 40)
        self.volume[
            loc[0] - self.t_size // 2 : loc[0] + self.t_size // 2,
            loc[1] - self.t_size // 2 : loc[1] + self.t_size // 2,
            loc[2] - self.t_size // 2 : loc[2] + self.t_size // 2,
        ] = vt.transform(
            self.template,
            rotation=rotation,
            rotation_units="rad",
            rotation_order="rzxz",
            device="cpu",
        )

        tm = TemplateMatchingGPU(
            0,
            0,
            self.volume,
            self.template,
            self.mask,
            self.angles,
            list(range(len(self.angles))),
            noise_correction=True,
        )
        score_volume, angle_volume, stats = tm.run()

        ind = np.unravel_index(score_volume.argmax(), self.volume.shape)
        self.assertTrue(
            score_volume.max() > 0.99, msg="lcc max value lower than expected"
        )
        self.assertEqual(angle_id, angle_volume[ind])
        self.assertSequenceEqual(loc, ind)
        expected_search_space = len(self.angles) * self.volume.size
        self.assertEqual(
            stats["search_space"],
            expected_search_space,
            msg="Search space should exactly equal this value",
        )
        self.assertAlmostEqual(
            stats["std"],
            0.005187,
            places=4,
            msg="Standard deviation of the search should be almost equal",
        )



import unittest
import pathlib
import warnings
import contextlib
from tempfile import TemporaryDirectory
import numpy as np
import mrcfile

from pytom_tm.io import read_mrc, read_mrc_meta_data, write_mrc, parse_relion5_star_data

FAILING_MRC = pathlib.Path(__file__).parent.joinpath(
    pathlib.Path("Data/human_ribo_mask_32_8_5.mrc")
)
# The below file was made with head -c 1024 human_ribo_mask_32_8_5.mrc > header_only.mrc
CORRUPT_MRC = pathlib.Path(__file__).parent.joinpath(
    pathlib.Path("Data/header_only.mrc")
)
RELION5_TOMOGRAMS_STAR = pathlib.Path(__file__).parent.joinpath(
    "Data/relion5_project_example/Tomograms/job009/tomograms.star"
)


class TestBrokenMRC(unittest.TestCase):
    def setUp(self):
        # Mute the RuntimeWarnings comming from other code-base inside these tests
        # following this SO answer: https://stackoverflow.com/a/45809502
        stack = contextlib.ExitStack()
        _ = stack.enter_context(warnings.catch_warnings())
        warnings.simplefilter("ignore")
        # The follwing line is better, but only works in python >= 3.11
        # _ = stack.enter_context(warnings.catch_warnings(action="ignore"))

        self.addCleanup(stack.close)

        # prep temporary directory
        tempdir = TemporaryDirectory()
        self.tempdirname = tempdir.name
        self.addCleanup(tempdir.cleanup)

    def test_read_mrc_minor_broken(self):
        # Test if this mrc can be read and if the approriate logs are printed
        with self.assertLogs(level="WARNING") as cm:
            mrc = read_mrc(FAILING_MRC)
        self.assertIsNotNone(mrc)
        self.assertEqual(len(cm.output), 1)
        self.assertIn(FAILING_MRC.name, cm.output[0])
        self.assertIn("make sure this is correct", cm.output[0])

    def test_read_mrc_too_broken(self):
        # Test if this mrc raises an error as expected
        with self.assertRaises(ValueError) as err:
            _ = read_mrc(CORRUPT_MRC)
        self.assertIn(CORRUPT_MRC.name, str(err.exception))
        self.assertIn("too corrupt", str(err.exception))

    def test_read_mrc_meta_data(self):
        # Test if this mrc can be read and if the approriate logs are printed
        with self.assertLogs(level="WARNING") as cm:
            mrc = read_mrc_meta_data(FAILING_MRC)
        self.assertIsNotNone(mrc)
        self.assertEqual(len(cm.output), 1)
        self.assertIn(FAILING_MRC.name, cm.output[0])
        self.assertIn("make sure this is correct", cm.output[0])

    def test_half_precision_read_write_cycle(self):
        array = np.random.rand(27).reshape((3, 3, 3)).astype(np.float16)
        fname = pathlib.Path(self.tempdirname) / "test_half.mrc"
        # Make sure no warnings are raised
        with self.assertNoLogs(level="WARNING"):
            write_mrc(fname, array, 1.0)
        # Make sure the file can be read back
        # make sure mode is as expected for float16
        # https://mrcfile.readthedocs.io/en/stable/source/mrcfile.html#mrcfile.utils.dtype_from_mode
        mrc = mrcfile.open(fname)
        self.assertEqual(mrc.header.mode, 12)
        mrc.close()
        # make sure dtype is expected
        mrc = read_mrc(fname)
        self.assertEqual(mrc.dtype, np.float16)
        # make sure data is identical
        np.testing.assert_equal(mrc, array)

    def test_cast_warning(self):
        # make sure a warning is raised when writing an integer based array
        array = np.random.rand(27).reshape((3, 3, 3)).astype(np.int32)
        fname = pathlib.Path(self.tempdirname) / "test_cast.mrc"
        with self.assertLogs(level="WARNING") as cm:
            write_mrc(fname, array, 1.0)
        self.assertEqual(len(cm.output), 1)
        self.assertIn("np.float32", cm.output[0])

    def test_parse_relion5_star_data(self):
        tomogram = pathlib.Path("rec_tomo200528_107.mrc")
        meta_data = parse_relion5_star_data(RELION5_TOMOGRAMS_STAR, tomogram)
        self.assertEqual(len(meta_data), 5)
        self.assertIsInstance(meta_data[0], float)
        self.assertIsInstance(meta_data[1], list)
        self.assertIsInstance(meta_data[2], list)
        self.assertIsInstance(meta_data[3], list)
        self.assertIsInstance(meta_data[3][0], dict)
        self.assertIsInstance(meta_data[4], int)

        tomogram = pathlib.Path("tomogram.mrc")
        with self.assertRaises(
            ValueError, msg="Unmatching tomograms name should raise an error."
        ):
            parse_relion5_star_data(RELION5_TOMOGRAMS_STAR, tomogram)

        tomogram = pathlib.Path("rec_tomogram200528_1077.mrc")
        with self.assertRaises(
            ValueError, msg="Partially matching tomogram name should raise an error."
        ):
            parse_relion5_star_data(RELION5_TOMOGRAMS_STAR, tomogram)

        tomogram = pathlib.Path("rec_tomogram200528_10.mrc")
        with self.assertRaises(
            ValueError, msg="Partially matching tomogram name should raise an error."
        ):
            parse_relion5_star_data(RELION5_TOMOGRAMS_STAR, tomogram)



import logging
from importlib import metadata

__version__ = metadata.version("pytom-match-pick")


try:
    import cupy  # noqa: F401
except (ModuleNotFoundError, ImportError):
    logging.warning(
        "Error for template matching: cupy installation not found or not functional."
    )



import pathlib
from pytom_tm.io import read_defocus_file, read_dose_file, read_tlt_file


# Dose and ctf params for tomo_104
cs = 2.7
amp = 0.08
vol = 200
defocus_data = read_defocus_file(
    pathlib.Path(__file__).parent.joinpath("Data").joinpath("test_imod.defocus")
)
CTF_PARAMS = []
for d in defocus_data:
    CTF_PARAMS.append(
        {
            "defocus": d,
            "amplitude_contrast": amp,
            "voltage": vol,
            "spherical_aberration": cs,
            "phase_shift_deg": 0.0,
        }
    )

ACCUMULATED_DOSE = read_dose_file(
    pathlib.Path(__file__).parent.joinpath("Data").joinpath("test_dose.txt")
)
TILT_ANGLES = read_tlt_file(
    pathlib.Path(__file__).parent.joinpath("Data").joinpath("test_angles.rawtlt")
)



import unittest
import numpy as np
from scipy.ndimage import center_of_mass
from pytom_tm.template import generate_template_from_map, phase_randomize_template


class TestTemplate(unittest.TestCase):
    def setUp(self):
        self.template = np.zeros((13, 13, 13), dtype=np.float32)
        self.template[2:5, 2:5, 7:9] = -1
        self.template[3:5, 3:5, 7:9] = 2
        self.template_center = (6, 6, 6)

    def test_template_padding(self):
        uneven_box = np.zeros((13, 13, 7))
        new_template = generate_template_from_map(uneven_box, 1, 1)
        self.assertEqual(
            new_template.shape, self.template.shape, msg="Box should be made square"
        )
        new_template = generate_template_from_map(uneven_box, 1, 2, output_box_size=20)
        self.assertEqual(
            new_template.shape,
            (20,) * 3,
            msg="Template was not padded to output box size",
        )

        with self.assertLogs(level="WARNING") as cm:
            new_template = generate_template_from_map(
                uneven_box, 1, 2, output_box_size=3
            )
        self.assertEqual(len(cm.output), 1)
        self.assertIn("Could not set specified box size", cm.output[0])

    def test_template_centering(self):
        new_template = generate_template_from_map(
            self.template,
            1,
            1,
            center=False,
        )
        square_sum = np.square(new_template - self.template).sum()
        self.assertTrue(
            square_sum < 10,
            msg="Template should not change strongly " "without recentering.",
        )
        new_template = generate_template_from_map(
            self.template,
            1,
            1,
            center=True,
        )
        square_sum = np.square(new_template - self.template).sum()
        self.assertTrue(square_sum > 10, msg="Template didnt change after shift")
        diff = np.array(center_of_mass(new_template**2)) - np.array(
            self.template_center
        )
        diff = np.abs(diff).sum()
        self.assertTrue(diff < 1, msg="Total shift difference should be small")
        # absolute of template should be identical due to square
        abs_template = generate_template_from_map(
            np.abs(self.template),
            1,
            1,
            center=True,
        )
        diff = np.array(center_of_mass(new_template**2)) - np.array(
            center_of_mass(abs_template**2)
        )
        diff = np.abs(diff).sum()
        self.assertTrue(diff < 1, msg="Absolute should provide exactly same center")

    def test_lowpass_resolution(self):
        # Test too low filter resolution
        with self.assertLogs(level="WARNING") as cm:
            _ = generate_template_from_map(
                self.template, 1.0, 1.0, filter_to_resolution=1.5
            )
        self.assertEqual(len(cm.output), 1)
        self.assertIn("Filter resolution", cm.output[0])
        self.assertIn("too low", cm.output[0])

        # Test working filter resolution
        with self.assertNoLogs(level="WARNING"):
            _ = generate_template_from_map(
                self.template, 1.0, 1.0, filter_to_resolution=2.5
            )

    def test_phase_randomize_template(self):
        randomized = phase_randomize_template(
            self.template,  # use default seed
        )
        self.assertEqual(self.template.shape, randomized.shape)
        self.assertGreater(
            (self.template != randomized).sum(),
            0,
            msg="After phase randomization the template should "
            "no longer be equal to the input.",
        )

        randomized_seeded = phase_randomize_template(
            self.template,
            11,  # use default seed
        )
        diff = np.abs(randomized_seeded - randomized).sum()
        self.assertNotEqual(
            diff, 0, msg="Different seed should return different randomization"
        )



import numpy as np
import unittest
from pytom_tm.weights import (
    create_wedge,
    create_ctf,
    create_gaussian_band_pass,
    radial_reduced_grid,
    radial_average,
    power_spectrum_profile,
    profile_to_weighting,
)
from testing_utils import TILT_ANGLES, ACCUMULATED_DOSE, CTF_PARAMS


class TestWeights(unittest.TestCase):
    def setUp(self):
        self.volume_shape_even = (10, 10, 10)
        self.volume_shape_uneven = (11, 11, 11)
        self.volume_shape_irregular = (7, 12, 6)
        self.voxel_size = 3.34
        self.low_pass = 10
        self.high_pass = 50

        self.reduced_even_shape_3d = (10, 10, 6)
        self.reduced_even_shape_2d = (10, 6)
        self.reduced_uneven_shape_3d = (11, 11, 6)
        self.reduced_uneven_shape_2d = (11, 6)
        self.reduced_irregular_shape_3d = (7, 12, 6 // 2 + 1)
        self.reduced_irregular_shape_2d = (7, 12 // 2 + 1)

    def test_radial_reduced_grid(self):
        with self.assertRaises(
            ValueError,
            msg="Radial reduced grid should raise ValueError if the shape is "
            "not 2- or 3-dimensional.",
        ):
            radial_reduced_grid((5,))
        with self.assertRaises(
            ValueError,
            msg="Radial reduced grid should raise ValueError if the shape is "
            "not 2- or 3-dimensional.",
        ):
            radial_reduced_grid((5,) * 4)

        self.assertEqual(
            radial_reduced_grid(self.volume_shape_even).shape,
            self.reduced_even_shape_3d,
            msg="3D radial reduced grid does not have the correct shape",
        )
        self.assertEqual(
            radial_reduced_grid(self.volume_shape_even[:2]).shape,
            self.reduced_even_shape_2d,
            msg="2D radial reduced grid does not have the correct shape",
        )

    def test_band_pass(self):
        with self.assertRaises(
            ValueError,
            msg="Bandpass should raise ValueError if both low and high pass are None",
        ):
            create_gaussian_band_pass(
                self.volume_shape_even, self.voxel_size, None, None
            )
        with self.assertRaises(
            ValueError,
            msg="Bandpass should raise ValueError if low pass resolution > high pass "
            "resolution",
        ):
            create_gaussian_band_pass(self.volume_shape_even, self.voxel_size, 50, 10)
        band_pass = create_gaussian_band_pass(
            self.volume_shape_even, self.voxel_size, self.low_pass, self.high_pass
        )
        low_pass = create_gaussian_band_pass(
            self.volume_shape_even, self.voxel_size, low_pass=self.low_pass
        )
        high_pass = create_gaussian_band_pass(
            self.volume_shape_even, self.voxel_size, high_pass=self.high_pass
        )

        self.assertEqual(
            band_pass.shape,
            self.reduced_even_shape_3d,
            msg="Bandpass filter does not have expected output shape",
        )
        self.assertEqual(
            band_pass.dtype,
            np.float64,
            msg="Bandpass filter does not have expected dtype",
        )
        self.assertEqual(
            low_pass.shape,
            self.reduced_even_shape_3d,
            msg="Low-pass filter does not have expected output shape",
        )
        self.assertEqual(
            low_pass.dtype,
            np.float64,
            msg="Low-pass filter does not have expected dtype",
        )
        self.assertEqual(
            high_pass.shape,
            self.reduced_even_shape_3d,
            msg="High-pass filter does not have expected output shape",
        )
        self.assertEqual(
            high_pass.dtype,
            np.float64,
            msg="High-pass filter does not have expected dtype",
        )

        self.assertTrue(
            np.sum((band_pass != low_pass) * 1) != 0,
            msg="Band-pass and low-pass should be different",
        )
        self.assertTrue(
            np.sum((band_pass != high_pass) * 1) != 0,
            msg="Band-pass and low-pass filter should be different",
        )
        self.assertTrue(
            np.sum((low_pass != high_pass) * 1) != 0,
            msg="Low-pass and high-pass filter should be different",
        )

    def test_create_wedge(self):
        with self.assertRaises(
            ValueError,
            msg="Create wedge should raise ValueError if tilt_angles list does not "
            "contain at least two values",
        ):
            create_wedge(self.volume_shape_even, [1.0], 1.0)
        with self.assertRaises(
            ValueError,
            msg="Create wedge should raise ValueError if tilt_angles input is not a "
            "list",
        ):
            create_wedge(self.volume_shape_even, 1.0, 1.0)
        with self.assertRaises(
            ValueError,
            msg="Create wedge should raise ValueError if voxel_size is smaller or "
            "equal to 0",
        ):
            create_wedge(self.volume_shape_even, TILT_ANGLES, 0.0)
        with self.assertRaises(
            ValueError,
            msg="Create wedge should raise ValueError if cut_off_radius is smaller or "
            "equal to 0",
        ):
            create_wedge(self.volume_shape_even, TILT_ANGLES, 1.0, cut_off_radius=0.0)

        # create test wedges
        structured_wedge = create_wedge(
            self.volume_shape_even,
            TILT_ANGLES,
            1.0,
            tilt_weighting=True,
            ctf_params_per_tilt=CTF_PARAMS,
        )
        symmetric_wedge = create_wedge(
            self.volume_shape_even,
            [TILT_ANGLES[0], TILT_ANGLES[-1]],
            1.0,
            tilt_weighting=False,
            ctf_params_per_tilt=CTF_PARAMS,
        )
        asymmetric_wedge = create_wedge(
            self.volume_shape_even,
            [TILT_ANGLES[0], TILT_ANGLES[-2]],
            1.0,
            tilt_weighting=False,
            ctf_params_per_tilt=CTF_PARAMS,
        )

        self.assertEqual(
            structured_wedge.shape,
            self.reduced_even_shape_3d,
            msg="Structured wedge does not have expected output shape",
        )
        self.assertEqual(
            structured_wedge.dtype,
            np.float32,
            msg="Structured wedge does not have expected dtype",
        )

        self.assertEqual(
            symmetric_wedge.shape,
            self.reduced_even_shape_3d,
            msg="Symmetric wedge does not have expected output shape",
        )
        self.assertEqual(
            symmetric_wedge.dtype,
            np.float32,
            msg="Symmetric wedge does not have expected dtype",
        )

        self.assertEqual(
            asymmetric_wedge.shape,
            self.reduced_even_shape_3d,
            msg="Asymmetric wedge does not have expected output shape",
        )
        self.assertEqual(
            asymmetric_wedge.dtype,
            np.float32,
            msg="Asymmetric wedge does not have expected dtype",
        )

        self.assertTrue(
            np.sum((symmetric_wedge != asymmetric_wedge) * 1) != 0,
            msg="Symmetric and asymmetric wedge should be different!",
        )

        structured_wedge = create_wedge(
            self.volume_shape_even,
            TILT_ANGLES,
            self.voxel_size,
            tilt_weighting=True,
            cut_off_radius=1.0,
            low_pass=self.low_pass,
            high_pass=self.high_pass,
        )
        self.assertEqual(
            structured_wedge.shape,
            self.reduced_even_shape_3d,
            msg="Wedge with band-pass does not have expected output shape",
        )
        self.assertEqual(
            structured_wedge.dtype,
            np.float32,
            msg="Wedge with band-pass does not have expected dtype",
        )

        # test shapes of wedges
        weights = create_wedge(
            self.volume_shape_even,
            TILT_ANGLES,
            self.voxel_size * 3,
            tilt_weighting=True,
            low_pass=40,
            accumulated_dose_per_tilt=ACCUMULATED_DOSE,
            ctf_params_per_tilt=CTF_PARAMS,
        )
        self.assertEqual(
            weights.shape,
            self.reduced_even_shape_3d,
            msg="3D CTF does not have the correct reduced fourier shape.",
        )
        weights = create_wedge(
            self.volume_shape_uneven,
            TILT_ANGLES,
            self.voxel_size * 3,
            tilt_weighting=True,
            low_pass=40,
            accumulated_dose_per_tilt=ACCUMULATED_DOSE,
            ctf_params_per_tilt=CTF_PARAMS,
        )
        self.assertEqual(
            weights.shape,
            self.reduced_uneven_shape_3d,
            msg="3D CTF does not have the correct reduced fourier shape.",
        )

        # test parameter flexibility of tilt_weighted wedge
        weights = create_wedge(
            self.volume_shape_even,
            TILT_ANGLES,
            self.voxel_size * 3,
            tilt_weighting=True,
            low_pass=self.low_pass,
            accumulated_dose_per_tilt=None,
            ctf_params_per_tilt=None,
        )
        self.assertEqual(
            weights.shape,
            self.reduced_even_shape_3d,
            msg="Tilt weighted wedge should also work without defocus and dose info.",
        )
        weights = create_wedge(
            self.volume_shape_even,
            TILT_ANGLES,
            self.voxel_size * 3,
            tilt_weighting=True,
            low_pass=self.low_pass,
            accumulated_dose_per_tilt=None,
            ctf_params_per_tilt=CTF_PARAMS[:1],
        )
        self.assertEqual(
            weights.shape,
            self.reduced_even_shape_3d,
            msg="Tilt weighted wedge should work with single defocus.",
        )

    def test_ctf(self):
        ctf_raw = create_ctf(
            self.volume_shape_even, self.voxel_size * 1e-10, 3e-6, 0.08, 300e3, 2.7e-3
        )
        ctf_cut = create_ctf(
            self.volume_shape_even,
            self.voxel_size * 1e-10,
            3e-6,
            0.08,
            300e3,
            2.7e-3,
            cut_after_first_zero=True,
        )
        self.assertEqual(
            ctf_raw.shape,
            self.reduced_even_shape_3d,
            msg="CTF does not have expected output shape",
        )
        self.assertTrue(
            np.sum((ctf_raw != ctf_cut) * 1) != 0,
            msg="CTF should be different when cutting it off after the first zero "
            "crossing",
        )

    def test_radial_average(self):
        x, y = 100, 50
        with self.assertRaises(
            ValueError,
            msg="Radial average should raise error if something other than 2d/3d "
            "array is provided.",
        ):
            radial_average(np.zeros(x))
        q, m = radial_average(np.zeros((x, y)))
        self.assertEqual(
            m.shape[0],
            x // 2 + 1,
            msg="Radial average shape should equal largest sampling dimension.",
        )
        q, m = radial_average(np.zeros((30, y)))
        self.assertEqual(
            m.shape[0],
            y,
            msg="Radial average shape should equal largest sampling dimension, "
            "considering Fourier reduced form.",
        )
        q, m = radial_average(np.zeros((20, x, y)))
        self.assertEqual(
            m.shape[0],
            x // 2 + 1,
            msg="Radial average shape should equal largest sampling dimension.",
        )
        q, m = radial_average(np.zeros((20, 30, y)))
        self.assertEqual(
            m.shape[0],
            y,
            msg="Radial average shape should equal largest sampling dimension, "
            "considering Fourier reduced form.",
        )

    def test_power_spectrum_profile(self):
        with self.assertRaises(
            ValueError,
            msg="Power spectrum profile should raise ValueError if input image is "
            "not 2- or 3-dimensional.",
        ):
            power_spectrum_profile(np.zeros(5))
        with self.assertRaises(
            ValueError,
            msg="Power spectrum profile should raise ValueError if input image is "
            "not 2- or 3-dimensional.",
        ):
            power_spectrum_profile(np.zeros((5,) * 4))
        profile = power_spectrum_profile(np.zeros(self.volume_shape_irregular))
        self.assertEqual(
            profile.shape,
            (max(self.volume_shape_irregular) // 2 + 1,),
            msg="Power spectrum profile output shape should be a 1-dimensional array "
            "with length equal to max(input_shape) // 2 + 1, corresponding to largest "
            "sampling component in Fourier space.",
        )

    def test_profile_to_weighting(self):
        with self.assertRaises(
            ValueError,
            msg="Profile to weighting should raise a ValueError if the profile is not "
            "1-dimensional.",
        ):
            profile_to_weighting(np.zeros((5, 5)), (5, 5))
        with self.assertRaises(
            ValueError,
            msg="Profile to weighting should raise a ValueError if the output shape "
            "for the weighting is not 2- or 3-dimensional.",
        ):
            profile_to_weighting(np.zeros(5), (5,))
        with self.assertRaises(
            ValueError,
            msg="Profile to weighting should raise a ValueError if the output shape "
            "for the weighting is not 2- or 3-dimensional.",
        ):
            profile_to_weighting(np.zeros(5), (5,) * 4)

        profile = power_spectrum_profile(np.zeros(self.volume_shape_irregular))
        self.assertEqual(
            profile_to_weighting(profile, self.volume_shape_irregular).shape,
            self.reduced_irregular_shape_3d,
            msg="Profile to weighting should return 3D Fourier reduced array.",
        )
        self.assertEqual(
            profile_to_weighting(profile, self.volume_shape_irregular[:2]).shape,
            self.reduced_irregular_shape_2d,
            msg="Profile to weighting should return 2D Fourier reduced array.",
        )



import numpy as np
import numpy.typing as npt
import scipy.ndimage as ndimage
import voltools as vt
from pytom_tm.io import UnequalSpacingError
from itertools import pairwise


constants = {
    # Dictionary of physical constants required for calculation.
    "c": 299792458,  # m/s
    "el": 1.60217646e-19,  # C
    "h": 6.62606896e-34,  # J*S
    "h_ev": 4.13566733e-15,  # eV*s
    "h_bar": 1.054571628e-34,  # J*s
    "h_bar_ev": 6.58211899e-16,  # eV*s
    "na": 6.02214179e23,  # mol-1
    "re": 2.817940289458e-15,  # m
    "rw": 2.976e-10,  # m
    "me": 9.10938215e-31,  # kg
    "me_ev": 0.510998910e6,  # ev/c^2
    "kb": 1.3806503e-23,  # m^2 kgs^-2 K^-1
    "eps0": 8.854187817620e-12,  # F/m
}


def hwhm_to_sigma(hwhm: float) -> float:
    """Convert half width of half maximum of a Gaussian to sigma by dividing by
    sqrt(2 * ln(2)).

    Parameters
    ----------
    hwhm: float
        half width of half maximum of Gaussian

    Returns
    -------
    sigma: float
        sigma of Gaussian
    """
    return hwhm / (np.sqrt(2 * np.log(2)))


def sigma_to_hwhm(sigma: float) -> float:
    """Convert sigma to half width of half maximum of a Gaussian by multiplying with
    sqrt(2 * ln(2)).

    Parameters
    ----------
    sigma: float
        sigma of Gaussian

    Returns
    -------
    hwhm: float
        half width of half maximum of Gaussian
    """
    return sigma * (np.sqrt(2 * np.log(2)))


def wavelength_ev2m(voltage: float) -> float:
    """Calculate wavelength of electrons from voltage.

    Parameters
    ----------
    voltage: float
        voltage of wave in eV

    Returns
    -------
    lambda: float
        wavelength of electrons in m
    """
    h = constants["h"]
    e = constants["el"]
    m = constants["me"]
    c = constants["c"]

    _lambda = h / np.sqrt(e * voltage * m * (e / m * voltage / c**2 + 2))

    return _lambda


def radial_reduced_grid(
    shape: tuple[int, int, int] | tuple[int, int], shape_is_reduced: bool = False
) -> npt.NDArray[float]:
    """Calculates a Fourier space radial reduced grid for the given input shape, with
    the 0 frequency in the center of the output image. Values range from 0 in the
    center to 1 at Nyquist frequency.

    By default, it is assumed shape belongs to a real space array, which causes the
    function to return a grid with the last dimension reduced, i.e. shape[-1] // 2 + 1
    (ideal for creating frequency dependent filters). However, setting
    radial_reduced_grid(..., shape_is_reduced=True) the shape is assumed to already be
    in a reduced form.

    Parameters
    ----------
    shape: Union[tuple[int, int, int], tuple[int, int]]
        2D/3D input shape, usually the .shape attribute of a numpy array
    shape_is_reduced: bool, default False
        whether the shape is already in a reduced fourier format, False by default

    Returns
    ----------
    radial_reduced_grid: npt.NDArray[float]
        fourier space frequency grid, 0 in center, 1 at nyquist
    """
    if len(shape) not in [2, 3]:
        raise ValueError("radial_reduced_grid() only works for 2D or 3D shapes")
    reduced_dim = shape[-1] if shape_is_reduced else shape[-1] // 2 + 1
    if len(shape) == 3:
        x = (
            np.abs(
                np.arange(
                    -shape[0] // 2 + shape[0] % 2, shape[0] // 2 + shape[0] % 2, 1.0
                )
            )
            / (shape[0] // 2)
        )[:, np.newaxis, np.newaxis]
        y = (
            np.abs(
                np.arange(
                    -shape[1] // 2 + shape[1] % 2, shape[1] // 2 + shape[1] % 2, 1.0
                )
            )
            / (shape[1] // 2)
        )[:, np.newaxis]
        z = np.arange(0, reduced_dim, 1.0) / (reduced_dim - 1)
        return np.sqrt(x**2 + y**2 + z**2)
    elif len(shape) == 2:
        x = (
            np.abs(
                np.arange(
                    -shape[0] // 2 + shape[0] % 2, shape[0] // 2 + shape[0] % 2, 1.0
                )
            )
            / (shape[0] // 2)
        )[:, np.newaxis]
        y = np.arange(0, reduced_dim, 1.0) / (reduced_dim - 1)
        return np.sqrt(x**2 + y**2)


def create_gaussian_low_pass(
    shape: tuple[int, int, int] | tuple[int, int],
    spacing: float,
    resolution: float,
) -> npt.NDArray[float]:
    """Create a 3D Gaussian low-pass filter with cutoff (or HWHM) that is reduced in
    fourier space.

    Parameters
    ----------
    shape: Union[tuple[int, int, int], tuple[int, int]]
        shape tuple with x,y or x,y,z dimension
    spacing: float
        voxel size in real space
    resolution: float
        resolution in real space to filter towards

    Returns
    ----------
    output: npt.NDArray[float]
        array containing the filter
    """
    q = radial_reduced_grid(shape)

    # 2 * spacing / resolution is cutoff in fourier space
    # then convert cutoff (hwhm) to sigma for gaussian function
    sigma_fourier = hwhm_to_sigma(2 * spacing / resolution)

    return np.fft.ifftshift(np.exp(-(q**2) / (2 * sigma_fourier**2)), axes=(0, 1))


def create_gaussian_high_pass(
    shape: tuple[int, int, int] | tuple[int, int],
    spacing: float,
    resolution: float,
) -> npt.NDArray[float]:
    """Create a 3D Gaussian high-pass filter with cutoff (or HWHM) that is reduced in
    fourier space.

    Parameters
    ----------
    shape: Union[tuple[int, int, int], tuple[int, int]]
        shape tuple with x,y or x,y,z dimension
    spacing: float
        voxel size in real space
    resolution: float
        resolution in real space to filter towards

    Returns
    ----------
    output: npt.NDArray[float]
        array containing the filter
    """
    q = radial_reduced_grid(shape)

    # 2 * spacing / resolution is cutoff in fourier space
    # then convert cutoff (hwhm) to sigma for gaussian function
    sigma_fourier = hwhm_to_sigma(2 * spacing / resolution)

    return np.fft.ifftshift(1 - np.exp(-(q**2) / (2 * sigma_fourier**2)), axes=(0, 1))


def create_gaussian_band_pass(
    shape: tuple[int, int, int] | tuple[int, int],
    spacing: float,
    low_pass: float | None = None,
    high_pass: float | None = None,
) -> npt.NDArray[float]:
    """Resolution bands presents the resolution shells where information needs to be
    maintained. For example the bands might be (150A, 40A). For a spacing of 15A
    (nyquist resolution is 30A) this is a mild low pass filter. However, quite some low
    spatial frequencies will be cut by it.

    Parameters
    ----------
    shape: Union[tuple[int, int, int], tuple[int, int]]
        shape tuple with x,y or x,y,z dimension
    spacing: float
        voxel size in real space
    low_pass: Optional[float], default None
        resolution of low-pass filter
    high_pass: Optional[float], default None
        resolution of high-pass filter

    Returns
    ----------
    output: npt.NDArray[float]
        array containing the band-pass filter
    """
    if high_pass is None and low_pass is None:
        raise ValueError("Either low-pass or high-pass needs to be set for band-pass")

    if high_pass is None:
        return create_gaussian_low_pass(shape, spacing, low_pass)
    elif low_pass is None:
        return create_gaussian_high_pass(shape, spacing, high_pass)
    elif low_pass >= high_pass:
        raise ValueError(
            "Second value of band-pass needs to be a high resolution shell."
        )
    else:
        q = radial_reduced_grid(shape)

        # 2 * spacing / resolution is cutoff in fourier space
        # then convert cutoff (hwhm) to sigma for gaussian function
        sigma_high_pass = hwhm_to_sigma(2 * spacing / high_pass)
        sigma_low_pass = hwhm_to_sigma(2 * spacing / low_pass)

        return np.fft.ifftshift(
            (1 - np.exp(-(q**2) / (2 * sigma_high_pass**2)))
            * np.exp(-(q**2) / (2 * sigma_low_pass**2)),
            axes=(0, 1),
        )


def create_wedge(
    shape: tuple[int, int, int],
    tilt_angles: list[float, ...],
    voxel_size: float,
    cut_off_radius: float = 1.0,
    angles_in_degrees: bool = True,
    low_pass: float | None = None,
    high_pass: float | None = None,
    tilt_weighting: bool = False,
    accumulated_dose_per_tilt: list[float, ...] | None = None,
    ctf_params_per_tilt: list[dict] | None = None,
) -> npt.NDArray[float]:
    """This function returns a wedge volume that is either symmetric or asymmetric
    depending on wedge angle input.

    Parameters
    ----------
    shape: tuple[int, int, int]
        real space shape of volume to which it needs to be applied
    tilt_angles: list[float, ...]
        tilt angles used for reconstructing the tomogram
    voxel_size: float
        voxel size is needed for the calculation of various filters
    cut_off_radius: float, default 1.
        cutoff as a fraction of nyquist, i.e. 1.0 means all the way to nyquist
    angles_in_degrees: bool, default True
        whether angles are in degrees or radians units
    low_pass: Optional[float], default None
        low pass filter resolution in A
    high_pass: Optional[float], default None
        high pass filter resolution in A
    tilt_weighting: bool, default False
        apply tilt weighting
    accumulated_dose_per_tilt: Optional[list[float, ...]], default None
        accumulated dose for each tilt for dose weighting
    ctf_params_per_tilt: Optional[list[dict]], default None
        ctf parameters for each tilt (see _create_tilt_weighted_wedge() for dict
        specification)

    Returns
    -------
    wedge: npt.NDArray[float]
        wedge volume that is a reduced fourier space object in z, i.e. shape[2] // 2 + 1
    """
    if not isinstance(tilt_angles, list) or len(tilt_angles) < 2:
        raise ValueError("Wedge generation needs at least a list of two tilt angles.")

    if voxel_size <= 0.0:
        raise ValueError(
            "Voxel size in create wedge is smaller or equal to 0, which is an invalid "
            "voxel spacing."
        )

    if cut_off_radius > 1:
        print(
            "Warning: wedge cutoff needs to be defined as a fraction of nyquist "
            "0 < c <= 1. Setting value to 1.0."
        )
        cut_off_radius = 1.0
    elif cut_off_radius <= 0:
        raise ValueError("Invalid wedge cutoff: needs to be larger than 0")

    if angles_in_degrees:
        tilt_angles_rad = [np.deg2rad(w) for w in tilt_angles]
    else:
        tilt_angles_rad = tilt_angles

    if tilt_weighting:
        if ctf_params_per_tilt is not None and len(ctf_params_per_tilt) == 1:
            # with only single defocus, copy the params for each tilt
            ctf_params_per_tilt = ctf_params_per_tilt * len(tilt_angles_rad)
        wedge = _create_tilt_weighted_wedge(
            shape,
            tilt_angles_rad,
            cut_off_radius,
            voxel_size,
            accumulated_dose_per_tilt=accumulated_dose_per_tilt,
            ctf_params_per_tilt=ctf_params_per_tilt,
        ).astype(np.float32)
    else:
        wedge_angles = (
            np.pi / 2 - np.abs(min(tilt_angles_rad)),
            np.pi / 2 - np.abs(max(tilt_angles_rad)),
        )
        if np.round(wedge_angles[0], 2) == np.round(wedge_angles[1], 2):
            wedge = _create_symmetric_wedge(
                shape, wedge_angles[0], cut_off_radius
            ).astype(np.float32)
        else:
            wedge = _create_asymmetric_wedge(
                shape, (wedge_angles[0], wedge_angles[1]), cut_off_radius
            ).astype(np.float32)
        if ctf_params_per_tilt is not None:
            # - take ctf params from approx. middle tilt as those are most accurate
            # - in case list has 1 element we take that
            ctf_params = ctf_params_per_tilt[len(ctf_params_per_tilt) // 2]
            wedge *= create_ctf(
                shape,
                voxel_size * 1e-10,
                **ctf_params,
            )

    if not (low_pass is None and high_pass is None):
        return wedge * create_gaussian_band_pass(
            shape, voxel_size, low_pass, high_pass
        ).astype(np.float32)
    else:
        return wedge


def _create_symmetric_wedge(
    shape: tuple[int, int, int], wedge_angle: float, cut_off_radius: float
) -> npt.NDArray[float]:
    """This function returns a symmetric wedge object.
    Function should not be imported, user should call create_wedge().

    Parameters
    ----------
    shape: tuple[int, int, int]
        real space shape of volume to which it needs to be applied
    wedge_angle: float
        angle describing symmetric wedge in radians
    cut_off_radius: float
        cutoff as a fraction of nyquist, i.e. 1.0 means all the way to nyquist

    Returns
    ----------
    wedge: npt.NDArray[float]
        wedge volume that is a reduced fourier space object in z, i.e. shape[2] // 2 + 1
    """
    x = (
        np.abs(
            np.arange(-shape[0] // 2 + shape[0] % 2, shape[0] // 2 + shape[0] % 2, 1.0)
        )
        / (shape[0] // 2)
    )[:, np.newaxis]
    z = np.arange(0, shape[2] // 2 + 1, 1.0) / (shape[2] // 2)

    # calculate the wedge mask with smooth edges
    wedge_2d = x - np.tan(wedge_angle) * z
    limit = (wedge_2d.max() - wedge_2d.min()) / (2 * min(shape[0], shape[2]) // 2)
    wedge_2d[wedge_2d > limit] = limit
    wedge_2d[wedge_2d < -limit] = -limit
    wedge_2d = (wedge_2d - wedge_2d.min()) / (wedge_2d.max() - wedge_2d.min())
    wedge_2d[shape[0] // 2 + 1, 0] = 1  # ensure that the zero frequency point equals 1

    # duplicate in x
    wedge = np.tile(wedge_2d[:, np.newaxis, :], (1, shape[1], 1))

    wedge[radial_reduced_grid(shape) > cut_off_radius] = 0

    # fourier shift to origin
    return np.fft.ifftshift(wedge, axes=(0, 1))


def _create_asymmetric_wedge(
    shape: tuple[int, int, int],
    wedge_angles: tuple[float, float],
    cut_off_radius: float,
) -> npt.NDArray[float]:
    """This function returns an asymmetric wedge object.
    Function should not be imported, user should call create_wedge().

    Parameters
    ----------
    shape: tuple[int, int, int]
        real space shape of volume to which it needs to be applied
    wedge_angles: tuple[float, float]
        two angles describing asymmetric missing wedge in radians
    cut_off_radius: float
        cutoff as a fraction of nyquist, i.e. 1.0 means all the way to nyquist

    Returns
    ----------
    wedge: npt.NDArray[float]
        wedge volume that is a reduced fourier space object in z, i.e. shape[2] // 2 + 1
    """
    x = (
        np.abs(
            np.arange(-shape[0] // 2 + shape[0] % 2, shape[0] // 2 + shape[0] % 2, 1.0)
        )
        / (shape[0] // 2)
    )[:, np.newaxis]
    z = np.arange(0, shape[2] // 2 + 1, 1.0) / (shape[2] // 2)

    # calculate wedge for first angle
    wedge_section = x - np.tan(wedge_angles[0]) * z
    limit = (wedge_section.max() - wedge_section.min()) / (
        2 * min(shape[0], shape[2]) // 2
    )
    wedge_section[wedge_section > limit] = limit
    wedge_section[wedge_section < -limit] = -limit
    wedge_section = (wedge_section - wedge_section.min()) / (
        wedge_section.max() - wedge_section.min()
    )

    # set top of the wedge
    wedge_2d = wedge_section.copy()

    # calculate wedge for second angle
    wedge_section = x - np.tan(wedge_angles[1]) * z
    limit = (wedge_section.max() - wedge_section.min()) / (
        2 * min(shape[0], shape[2]) // 2
    )
    wedge_section[wedge_section > limit] = limit
    wedge_section[wedge_section < -limit] = -limit
    wedge_section = (wedge_section - wedge_section.min()) / (
        wedge_section.max() - wedge_section.min()
    )

    # set bottom of wedge and put 0 frequency to 1
    wedge_2d[shape[0] // 2 + 1 :] = wedge_section[shape[0] // 2 + 1 :]
    wedge_2d[shape[0] // 2 + 1, 0] = 1

    # duplicate in x
    wedge = np.tile(wedge_2d[:, np.newaxis, :], (1, shape[1], 1))

    wedge[radial_reduced_grid(shape) > cut_off_radius] = 0

    # fourier shift to origin
    return np.fft.ifftshift(wedge, axes=(0, 1))


def _create_tilt_weighted_wedge(
    shape: tuple[int, int, int],
    tilt_angles: list[float, ...],
    cut_off_radius: float,
    pixel_size_angstrom: float,
    accumulated_dose_per_tilt: list[float, ...] | None = None,
    ctf_params_per_tilt: list[dict] | None = None,
) -> npt.NDArray[float]:
    """
    The following B-factor heuristic is used (as mentioned in the M paper, and
    introduced in RELION 1.4):
        "The B factor is increased by 4Å2 per 1e− Å−2 of exposure, and each tilt
        is weighted as cos θ."

    Relation between B-factor and the sigma of a gaussian:

        B = 8 * pi ** 2 * sigma_motion ** 2

    i.e. sigma_motion = sqrt( B / (8 * pi ** 2)). Belonging to a Gaussian blur:

        exp( -2 * pi ** 2 * sigma_motion ** 2 * q ** 2 )

    Parameters
    ----------
    shape: tuple[int, int, int]
        shape of volume to model the wedge for
    tilt_angles: list[float, ...]
        tilt angles is a list of angle in radian units
    cut_off_radius: float
        cut off for the mask as a fraction of nyquist, value between 0 and 1
    pixel_size_angstrom: float
        the pixel size as a value in Å
    accumulated_dose_per_tilt: list[float, ...], default None
        the accumulated dose in e− Å−2
    ctf_params_per_tilt: list[dict, ...], default None
        the ctf parameters per tilt angle, list of dicts
        where each dict has the following keys:
        - 'defocus'; in um
        - 'amplitude'; fraction of amplitude contrast between 0 and 1
        - 'voltage'; in keV
        - 'cs'; spherical abberation in mm
        - 'phase_shift_deg'; phase shift for phase plates in deg

    Returns
    -------
    wedge: npt.NDArray[float]
        structured wedge mask in fourier reduced form, i.e. output shape is
        (shape[0], shape[1], shape[2] // 2 + 1)
    """
    if accumulated_dose_per_tilt is not None and len(accumulated_dose_per_tilt) != len(
        tilt_angles
    ):
        raise ValueError(
            "in _create_tilt_weighted_wedge the list of accumulated dose per tilt does "
            "not have the same length as the tilt angle list!"
        )
    if ctf_params_per_tilt is not None and len(ctf_params_per_tilt) != len(tilt_angles):
        raise ValueError(
            "in _create_tilt_weighted_wedge the list of CTF parameters per tilt does "
            "not have the same length as the tilt angle list!"
        )
    if not all([shape[0] == s for s in shape[1:]]):
        raise UnequalSpacingError(
            "Input shape for structured wedge needs to be a square box. "
            "Otherwise the frequencies in fourier space are not equal across "
            "dimensions."
        )

    image_size = shape[0]  # assign to size variable as all dimensions are equal size
    tilt = np.zeros(shape)
    q_grid = radial_reduced_grid(shape)
    tilt_weighted_wedge = np.zeros((image_size, image_size, image_size // 2 + 1))

    # create ramp weights to correct tilt summation for overlap
    tilt_increment = min([abs(x - y) for x, y in pairwise(tilt_angles)])
    # Crowther freq. determines till what point adjacent tilts overlap in Fourier space
    overlap_frequency = 1 / (tilt_increment * image_size)
    freq_1d = (
        np.abs(
            np.arange(
                -image_size // 2 + image_size % 2, image_size // 2 + image_size % 2, 1.0
            )
        )
        / (image_size // 2)
        * 0.5
    )  # multiply with .5 for nyquist frequency
    ramp_filter = freq_1d / overlap_frequency
    ramp_filter[ramp_filter > 1] = 1  # linear increase up to overlap frequency

    # generate 2d weights along the tilt axis
    ramp_weighting = np.tile(ramp_filter[:, np.newaxis], (1, image_size))

    for i, alpha in enumerate(tilt_angles):
        if ctf_params_per_tilt is not None:
            ctf = np.fft.fftshift(
                create_ctf(
                    (image_size,) * 2,
                    pixel_size_angstrom * 1e-10,
                    **ctf_params_per_tilt[i],
                ),
                axes=0,
            )
            tilt[:, :, image_size // 2] = (
                np.concatenate(
                    (  # duplicate and flip the CTF around the 0 frequency;
                        # then concatenate to make it non-reduced
                        np.flip(ctf[:, 1 : 1 + image_size - ctf.shape[1]], axis=1),
                        ctf,
                    ),
                    axis=1,
                )
                * ramp_weighting
            )
        else:
            tilt[:, :, image_size // 2] = ramp_weighting

        # rotate the image weights to the tilt angle
        rotated = np.flip(
            vt.transform(
                tilt,
                rotation=(0, alpha, 0),
                rotation_units="rad",
                rotation_order="rxyz",
                center=(image_size // 2,) * 3,
                interpolation="filt_bspline",
                device="cpu",
            )[:, :, : image_size // 2 + 1],  # crop back z-axis to reduced Fourier form
            axis=2,
        )

        # weight with exposure and tilt dampening
        if accumulated_dose_per_tilt is not None:
            q_squared = (q_grid / (2 * pixel_size_angstrom)) ** 2
            sigma_motion = np.sqrt(accumulated_dose_per_tilt[i] * 4 / (8 * np.pi**2))
            weighted_tilt = (
                rotated
                * np.cos(alpha)  # apply tilt-dependent weighting
                * np.exp(
                    -2 * np.pi**2 * sigma_motion**2 * q_squared
                )  # apply dose-weighting
            )
        else:
            weighted_tilt = (
                rotated * np.cos(alpha)  # apply tilt-dependent weighting
            )

        tilt_weighted_wedge += weighted_tilt

    tilt_weighted_wedge[q_grid > cut_off_radius] = 0

    return np.fft.ifftshift(tilt_weighted_wedge, axes=(0, 1))


def create_ctf(
    shape: tuple[int, int, int] | tuple[int, int],
    pixel_size: float,
    defocus: float,
    amplitude_contrast: float,
    voltage: float,
    spherical_aberration: float,
    cut_after_first_zero: bool = False,
    flip_phase: bool = False,
    phase_shift_deg: float = 0.0,
) -> npt.NDArray[float]:
    """Create a CTF in a 3D volume in reduced format.

    Parameters
    ----------
    shape: Union[tuple[int, int, int], tuple[int, int]]
        dimensions of volume to create ctf in
    pixel_size: float
        pixel size for ctf in m
    defocus: float
        defocus for ctf in m
    amplitude_contrast: float
        the fraction of amplitude contrast in the ctf
    voltage: float
        acceleration voltage of the microscope in eV
    spherical_aberration: float
        spherical aberration in m
    cut_after_first_zero: bool, default False
        whether to cut ctf after first zero crossing
    flip_phase: bool, default False
        make ctf fully positive/negative to imitate ctf correction by phase flipping
    phase_shift_deg: float, default .0
        additional phase shift to model phase plates, similar to
        `https://github.com/dtegunov/tom_deconv` except the ctf defintion in tom
        produces the inverse curve of what we have here

    Returns
    -------
    ctf: npt.NDArray[float]
        CTF in 3D
    """
    k = radial_reduced_grid(shape) / (2 * pixel_size)  # frequencies in fourier space

    _lambda = wavelength_ev2m(voltage)

    # phase contrast transfer
    chi = (
        np.pi * _lambda * defocus * k**2
        - 0.5 * np.pi * spherical_aberration * _lambda**3 * k**4
    )
    # amplitude contrast term
    tan_term = np.arctan(amplitude_contrast / np.sqrt(1 - amplitude_contrast**2))

    # determine the ctf
    ctf = -np.sin(chi + tan_term + np.deg2rad(phase_shift_deg))

    if cut_after_first_zero:  # find frequency to cut first zero

        def chi_1d(q):
            return (
                np.pi * _lambda * defocus * q**2
                - 0.5 * np.pi * spherical_aberration * _lambda**3 * q**4
            )

        def ctf_1d(q):
            return -np.sin(chi_1d(q) + tan_term)

        # sample 1d ctf and get indices of zero crossing
        k_range = np.arange(max(k.shape)) / max(k.shape) / (2 * pixel_size)
        values = ctf_1d(k_range)
        zero_crossings = np.where(np.diff(np.sign(values)))[0]

        # for overfocus the first crossing needs to be skipped,
        # for example see: Yonekura et al. 2006 JSB
        k_cutoff = (
            k_range[zero_crossings[0]] if defocus > 0 else k_range[zero_crossings[1]]
        )

        # filter the ctf with the cutoff frequency
        ctf[k > k_cutoff] = 0

    if flip_phase:  # take absolute, ensures matching contrast
        ctf = np.abs(ctf)
    else:  # multiply the ctf with -1 if we have overfocus, this allows the user to
        # always match the contrast of the input template with the contrast of the
        # tomogram: if the tomogram is black the reference should be black.
        ctf *= -1 if defocus > 0 else 1

    return np.fft.ifftshift(ctf, axes=(0, 1) if len(shape) == 3 else 0)


def radial_average(
    weights: npt.NDArray[float],
) -> tuple[npt.NDArray[float], npt.NDArray[float]]:
    """This calculates the radial average of a reduced fourier space function.

    Parameters
    ----------
    weights: npt.NDArray[float]
        3D array to be radially averaged: in fourier reduced form and with the origin
        in the corner.

    Returns
    -------
    (q, mean): tuple[npt.NDArray[float], npt.NDArray[float]]
        A tuple of two 1d numpy arrays. Their length equals half of largest
        input dimension.
    """
    if len(weights.shape) not in [2, 3]:
        raise ValueError("Radial average calculation only works for 2d/3d arrays")

    # get the number of sampling points from the largest fourier dimension,
    # unless the reduced dimensions is already the largest one
    sampling_points = max(max(weights.shape[:-1]) // 2 + 1, weights.shape[-1])

    q = np.arange(sampling_points)
    q_grid = np.floor(
        # convert to radial indices in the fourier power spectrum,
        # 0.5 is added to obtain the correct ring
        radial_reduced_grid(weights.shape, shape_is_reduced=True)
        * (sampling_points - 1)
        + 0.5
    ).astype(int)
    mean = ndimage.mean(
        np.fft.fftshift(weights, axes=(0, 1) if len(weights.shape) == 3 else 0),
        labels=q_grid,
        index=q,
    )

    return q, mean


def power_spectrum_profile(image: npt.NDArray[float]) -> npt.NDArray[float]:
    """Calculate the power spectrum for a real space array and then find the profile
    (radial average).

    Parameters
    ----------
    image: npt.NDArray[float]
        2D/3D array in real space for which the power spectrum profile needs to be
        calculated

    Returns
    -------
    profile: npt.NDArray[float]
        A 1D numpy array
    """
    if len(image.shape) not in [2, 3]:
        raise ValueError(
            "Power spectrum profile calculation only works for 2d/3d arrays."
        )

    _, power_profile = radial_average(np.abs(np.fft.rfftn(image)) ** 2)

    return power_profile


def profile_to_weighting(
    profile: npt.NDArray[float], shape: tuple[int, int] | tuple[int, int, int]
) -> npt.NDArray[float]:
    """Calculate a radial weighing (filter) from a spectrum profile.

    Parameters
    ----------
    profile: npt.NDArray[float]
        power spectrum profile (or other 1d profile) to transform in a fourier space
        filter
    shape: Union[tuple[int, int], tuple[int, int, int]]
        2D/3D array shape in real space for which the fourier reduced weights are
        calculated

    Returns
    -------
    weighting: npt.NDArray[float]
        Reduced Fourier space weighting for shape
    """
    if len(profile.shape) != 1:
        raise ValueError("Profile passed to profile_to_weighting is not 1-dimensional.")
    if len(shape) not in [2, 3]:
        raise ValueError("Shape passed to profile_to_weighting needs to be 2D/3D.")

    q_grid = radial_reduced_grid(shape)

    weights = ndimage.map_coordinates(
        profile, q_grid.flatten()[np.newaxis, :] * profile.shape[0], order=1
    ).reshape(q_grid.shape)

    weights[q_grid > 1] = 0

    return np.fft.ifftshift(weights, axes=(0, 1) if len(shape) == 3 else 0)
